import numpy as np
from scipy.stats import qmc, beta
import tqdm
from typing import Callable, Union
from pyuncertainnumber.propagation.utils import Propagation_results


def index_to_bool_(index: np.ndarray, dim=2):
    """Converts a vector of indices to an array of boolean pairs for masking.

    args:
        index: A NumPy array of integer indices representing selected elements or categories.
                The values in `index` should be in the range [0, dim-1].
        dim (scalar): The number of categories or dimensions in the output boolean array. Defaults to 2.

    signature:
      index_to_bool_(index:np.ndarray,dim=2) -> tuple

    note:
          - the augument `index` is an np.ndaray of the index of intervals.
          - the argument `dim` will specify the function mapping of variables to be propagated.
          - If dim > 2,  e.g. (2,0,1,0) the array of booleans is [(0,0,1),(1,0,0),(0,1,0),(1,0,0)].

    returns:
      A NumPy array of boolean pairs representing the mask.
    """

    index = np.asarray(index, dtype=int)
    return np.asarray([index == j for j in range(dim)], dtype=bool)


def sampling_method(
    x: np.ndarray,
    f: Callable,
    results: Propagation_results = None,
    method="monte_carlo",
    b: Union[float, np.ndarray] = 1.0,
    n_sam: int = 500,
    endpoints=False,
    save_raw_data="no",
) -> Propagation_results:  # Specify return type
    """sampling of intervals

    args:
        x (np.ndarray): A 2D NumPy array where each row represents an input variable and
                            the two columns define its lower and upper bounds (interval).
        f (Callable): A callable function that takes a 1D NumPy array of input values and returns the
                        corresponding output(s). Can be None, in which case only samples are generated.
        results (Propagation_results, optional): An object to store propagation results.
                                            Defaults to None, in which case a new
                                            `Propagation_results` object is created.
        method (str, optional): The sampling method to use. Choose from:
                                 - 'monte_carlo': Monte Carlo sampling (random sampling from uniform distributions)
                                 - 'latin_hypercube': Latin Hypercube sampling (stratified sampling for better space coverage)
                                Defaults to 'monte_carlo'.
        b (float or np.ndarray): The shape parameter(s) for the bathtub distribution(s).
                                  If a float, the same shape parameter is used for all inputs.
                                  If an np.ndarray, it should have the same length as the number of input variables,
                                  providing a separate shape parameter for each input. Default: 1.0 (uniform distribution).
        n_sam (int): The number of samples to generate for the chosen sampling method.
        endpoints (bool, optional): If True, include the interval endpoints in the sampling.
                                    Defaults to False.
        save_raw_data (str, optional): Acts as a switch to enable or disable the storage of raw input data when a function (f) 
          is not provided.
          - 'no': Returns an error that no function is provided.
          - 'yes': Returns the full arrays of unique input combinations.

    signature:
        sampling_method(x:np.ndarray, f:Callable, results:Propagation_results = None,  method ='montecarlo', b=1, n_sam:int,
        endpoints=False,  save_raw_data = 'no') -> Propagation_results

    note:
        - The `f` function may return multiple outputs.
        - For the Monte carlo method, sampling with bathtub distributions allows for more flexible exploration of the input space
          compared to uniform distributions.
        - Bathtub distributions are not directly applicable to the latin hypercube which samples from uniform input distributions. 
        - To account for the extreme values of each interval, sampling methods are combined with the endpoints method. 
        - Overall, the sampling method is expected to yield non-conservative results if used to propagate input
          intervals through a black-box model.

    returns:
       `Propagation_results` object(s) containing:
            - 'un': UncertainNumber object(s) to characterise the interval(s) of the output(s).
            - 'raw_data' (dict): Dictionary containing raw data shared across output(s):
                    - 'x' (np.ndarray): Input values.
                    - 'f' (np.ndarray): Output values.
                    - 'min' (np.ndarray): Array of dictionaries, one for each output,
                              containing 'f' and corresponding 'x' for the minimum of that output.
                    - 'max' (np.ndarray): Array of dictionaries, one for each output,
                              containing 'f' and corresponding 'x' for the maximum of that output.
                    - 'bounds' (np.ndarray): 2D array of lower and upper bounds for each output.

    raises:
        ValueError if no function is provided and save_raw_data is 'no'.

    example:
        >>> x = np.array([[1, 2], [3, 4], [5, 6]])  # Define input intervals
        >>> f = lambda x: x[0] + x[1] + x[2]  # Define the function
        >>> y = sampling_method(x, f,  method='monte_carlo', b=1,  n_sam=500, endpoints=False, save_raw_data='no')
    """
    if results is None:
        results = Propagation_results()  # Create an instance of Propagation_results

    m = x.shape[0]
    lo = x[:, 0]
    hi = x[:, 1]

    # Ensure b is an array of the correct length
    if isinstance(b, float):
        b_arr = np.full(m, b)
    else:
        b_arr = b
        
    if method == "monte_carlo":
        X = np.zeros((n_sam, m))
        for i in range(m):
            X[:, i] = lo[i] + (hi[i] - lo[i]) * beta.rvs(1/b_arr[i], 1/b_arr[i], size=n_sam)
    elif method == "latin_hypercube":
        sampler = qmc.LatinHypercube(m)
        X = lo + (hi - lo) * sampler.random(n_sam)
    else:
        raise ValueError(
            "Invalid sampling method. Choose 'monte_carlo' or 'latin_hypercube'."
        )

    if endpoints:
        m = x.shape[0]
        Total = (2**m )  # Total number of endpoint combination for the give x input variables
        # Initialize array for endpoint combinations
        X_end = np.zeros((Total, m))
        for j in range(Total):
            # tuple of 0s and 1s
            index = tuple([j // 2**h - (j // 2 ** (h + 1)) * 2 for h in range(m)])
            itb = index_to_bool_(index).T
            X_end[j, :] = x[itb]
        # Combine generated samples with endpoint combinations
        X = np.vstack([X, X_end])

    if f is not None:  # Only evaluate if f is provided
        all_output = np.array([f(xi) for xi in tqdm.tqdm(X, desc="Function evaluations")])

        try:
            num_outputs = len(all_output[0])
        except TypeError:
            num_outputs = 1  # If f returns a single value

        # Reshape all_output to a 2D array (Corrected)
        all_output = np.array(all_output).reshape(-1, num_outputs)

        # Calculate lower and upper bounds for each output
        lower_bound = np.min(all_output, axis=0)
        upper_bound = np.max(all_output, axis=0)

        # Find indices of minimum and maximum values for each output
        min_indices = np.array(
            [
                np.where(all_output[:, i] == lower_bound[i])[0]
                for i in range(num_outputs)
            ]
        )
        max_indices = np.array(
            [
                np.where(all_output[:, i] == upper_bound[i])[0]
                for i in range(num_outputs)
            ]
        )

        # Populate the results object (Corrected)
        for i in range(num_outputs):
            results.raw_data["min"] = np.append(
                results.raw_data["min"], {"x": X[min_indices[i]], "f": lower_bound[i]}
            )
            results.raw_data["max"] = np.append(
                results.raw_data["max"], {"x": X[max_indices[i]], "f": upper_bound[i]}
            )
            results.raw_data["bounds"] = (
                np.vstack(
                    [
                        results.raw_data["bounds"],
                        np.array([lower_bound[i], upper_bound[i]]),
                    ]
                )
                if results.raw_data["bounds"].size
                else np.array([lower_bound[i], upper_bound[i]])
            )

        # if save_raw_data == 'yes': (This part is already handled correctly)
        results.raw_data["f"] = all_output
        results.raw_data["x"] = X

    elif save_raw_data == "yes":  # If f is None and save_raw_data is 'yes'
        results.add_raw_data(x=X)

    else:
        print(
            "No function is provided. Select save_raw_data = 'yes' to save the input combinations"
        )

    return results


# # Example usage with different parameters for minimization and maximization
# f = lambda x: x[0] + x[1] + x[2]  # Example function

# # Determine input parameters for function and method
# x_bounds = np.array([[1, 2], [3, 4], [5, 6]])
# n=20
# # Call the method
# y = sampling_method(x_bounds,f=f,  method  ='monte_carlo' , n_sam=n, endpoints= False, save_raw_data = 'yes')
# print(y.raw_data['bounds'])



from .bo import BayesOpt
from .ga import GA


def get_range_BO(
    f, dimension, xc_bounds, acquisition_function, verbose=False, **kwargs
):
    """get the range of the black-box function using BayesOpt

    note:
        for a less verbose output, use (convergence_curve=False, progress_bar=False)
    """

    if not verbose:  # quiet mode
        V = 0

    min_task = BayesOpt(
        f=f,
        dimension=dimension,
        task="minimisation",
        xc_bounds=xc_bounds,
        acquisition_function=acquisition_function,
        **kwargs,
    )
    min_task.run(verbose=V)
    min = min_task.optimal_target

    max_task = BayesOpt(
        f=f,
        dimension=dimension,  # type: ignore
        task="maximisation",
        xc_bounds=xc_bounds,
        acquisition_function=acquisition_function,
        **kwargs,
    )
    max_task.run(verbose=V)
    max = max_task.optimal_target
    return min, max


def get_range_GA(f, dimension, varbound, algorithm_param=None, verbose=False, **kwargs):
    """get the range of the black-box function using GA"""

    if not verbose:
        kwargs["convergence_curve"] = False
        kwargs["progress_bar"] = False

    min_task = GA(f, task="minimisation", dimension=dimension, varbound=varbound)
    min_task.run(algorithm_param=algorithm_param, **kwargs)
    min = min_task.optimal_target

    max_task = GA(f, task="maximisation", dimension=dimension, varbound=varbound)
    max_task.run(algorithm_param=algorithm_param, **kwargs)
    max = max_task.optimal_target

    return min, max



from __future__ import annotations
from typing import TYPE_CHECKING
import matplotlib.pyplot as plt

if TYPE_CHECKING:
    from .pba.pbox_abc import Pbox
    from .pba.dss import DempsterShafer


def env_helper(elements: list, env):
    """help visualise the envelope"""

    fig, ax = plt.subplots()

    for p in elements:
        p.plot(ax=ax, zorder=20)

    env.plot(fill_color="salmon", ax=ax, bound_colors=["black", "black"], zorder=50)



import numpy as np
from scipy.stats import (
    bernoulli,
    beta,
    betabinom,
    binom,
    chi2,
    expon,
    gamma,
    norm,
    poisson,
    uniform,
    t,
    gumbel_r,
    lognorm,
    laplace,
    logistic,
    pareto,
    powerlaw,
    triang,
    geom,
    loguniform,
    f as F,
)
import scipy
import scipy.stats as sps
from ..pba.intervals.intervalOperators import mean
from ..pba.intervals import Interval
from pyuncertainnumber import pba
from ..pba.distributions import Distribution as D
from ..pba.distributions import named_dists
from ..pba.pbox_parametric import named_pbox
import functools
from .core import makeUN


""" Here we define the statistical inference functions from data for the UncertainNumber class. """


@makeUN
def fit(method: str, family: str, data: np.ndarray):
    """parametric estimator to fit a distribution from data

    args:
        - method (str): method of fitting, e.g., {'mle' or 'mom'} 'entropy', 'pert', 'fermi', 'bayesian'
        - family (str): distribution family to be fitted
        - data (np.ndarray): data to be fitted

    note:
        - supported family list can be found in xx.

    return:
        - the return from the constructors below are `scipy.stats.dist` objects or `UN` objects depending on the decorator

    example:
        >>> pun.fit('mle', 'norm', np.random.normal(0, 1, 100))
    """
    match method:
        case "mle":
            try:
                return mle.get(family)(data)
            except:
                return smle.get(family)(data)
        case "mom":
            return mom.get(family)(data)
        case _:
            raise ValueError("method not supported")


def makedist(shape: str):
    """a decorator to transform a `sps.dist` into `Distribution` objects"""

    def decorator_make_dist(func):
        @functools.wraps(func)
        def wrapper_decorator(*args, **kwargs):  # input array x
            return D.dist_from_sps(func(*args, **kwargs), shape)

        return wrapper_decorator

    return decorator_make_dist


# TODO: tested with expon which seems incorrect.
# TODO: stats.expon takes `1/lambda` as the scale, which is the inverse of the rate parameter.
def singleParamPattern(x, shape: str):
    if isinstance(x, sps.CensoredData | np.ndarray | list):
        return D.dist_from_sps(named_dists.get(shape)(mean(x)), shape=shape)
    elif isinstance(x, Interval):
        return named_pbox.get(shape)(mean(x))
    else:
        raise TypeError("Input data type not supported")


###############################################################################
# Method-of-Moment distribution constructors (matching central moments of x)
###############################################################################
""" hint: given interval data x, return pbox """


def MMbernoulli(x):
    """a first attempt to Maximum likelihood estimation for Bernoulli distribution
        which accepts both precise and imprecise data;

    #! the example of `singleparam` pattern
    #! to change, add the 'interval_measurement' decorator
    note:
        the attempt is successful per se, but not accommodating to the top-level calling signature yet.

        - precise data returns precise distrubution
        - imprecise data need to be in Interval type to return a pbox
        - interval data can return either a precise distribution or a pbox
    """
    if isinstance(x, sps.CensoredData | np.ndarray | list):
        return D.dist_from_sps(expon(*sps.bernoulli.fit(x)), shape="bernoulli")
    elif isinstance(x, Interval):
        return pba.bernoulli(mean(x))
    else:
        raise TypeError("Input data type not supported")


@makedist("beta")
def MMbeta(x: np.ndarray):
    m = x.mean()
    s = x.var()

    alpha = m * (m * (1 - m) / s - 1)
    beta_p = (1 - m) * (m * (1 - m) / s - 1)
    return beta(alpha, beta_p)


@makedist("betabinom")
def MMbetabinomial(n: int, x):  # **
    #! 'x**2' variable repetition
    # n must be provided; it's not estimated from data
    # https://en.wikipedia.org/wiki/Beta-binomial_distribution#Example:
    # MMbetabinomial(n=12,rep(0:12,c(3,24,104,286,670,1033,1343,1112,829,478,181,45,7)))
    m1 = x.mean()
    m2 = (x**2).mean()
    d = n * (m2 / m1 - m1 - 1) + m1
    return betabinom(n, (n * m1 - m2) / d, (n - m1) * (n - m2 / m1) / d)


@makedist("binom")
def MMbinomial(x, n: int):  # **
    #! the return seems overcomplicated?
    """
    args:
         - n (int): number of trials
    """
    return binom(n, mean() / n)
    # a = x.mean()
    # b = x.std()
    # return binom(int(np.abs(np.round(a/(1-b**2/a)))), np.abs(1-b**2/a))


@makedist("chi2")
def MMchisquared(x):
    return chi2(np.round(x.mean()))


# def MMchisquared(x):
#     #! TODO interval outward_rounding
#     if isinstance(x, sps.CensoredData | np.ndarray | list):
#         return D.dist_from_sps(chi2(outward_rounding(mean(x))), shape='chi2')
#     elif isinstance(x, Interval):
#         return pba.expon(mean(x))
#     else:
#         raise TypeError('Input data type not supported')


# TODO: this is incorrect as **kwargs are not passed.
# TODO: expon takes (scale=1/lambda) as kwargs
def MMexponential(x):
    return singleParamPattern(x, "expon")


@makedist("f")
def MMF(x):  # **
    w = 2 / (1 - 1 / x.mean())
    return F(
        np.round(
            (2 * w**3 - 4 * w**2) / ((w - 2) ** 2 * (w - 4) * x.std() ** 2 - 2 * w**2)
        ),
        np.round(w),
    )


@makedist("gamma")
def MMgamma(x):  # **
    a = x.mean()
    b = x.std()
    return gamma(b**2 / a, 1 / (a / b) ** 2)  # gamma1(a, b) ~ gamma(b²/a, (a/b)²)


@makedist("geometric")
def MMgeometric(x):
    pass


def MMgeometric(x):
    if isinstance(x, sps.CensoredData | np.ndarray | list):
        return D.dist_from_sps(geom(1 / (1 + mean(x))), shape="geometric")
    elif isinstance(x, Interval):
        return pba.geom(1 / (1 + mean(x)))
    else:
        raise TypeError("Input data type not supported")


def MMpascal(x):
    if isinstance(x, sps.CensoredData | np.ndarray | list):
        return D.dist_from_sps(geom(1 / (1 + mean(x))), shape="pascal")
    elif isinstance(x, Interval):
        return pba.geom(1 / (1 + mean(x)))
    else:
        raise TypeError("Input data type not supported")


# def MMgumbel0(x) : return(gumbel(x.mean() - 0.57721 * x.std() * np.sqrt(6)/ np.pi, x.std() * np.sqrt(6)/ np.pi))       #**  # https://stackoverflow.com/questions/51427764/using-method-of-moments-with-gumbel-r-in-python-scipy-stats-gumbel-r


@makedist("gumbel")
def MMgumbel(x):  # **
    # https://stackoverflow.com/questions/51427764/using-method-of-moments-with-gumbel-r-in-python-scipy-stats-gumbel-r
    scale = np.sqrt(6) / np.pi * np.std(x)
    loc = np.mean(x) - np.euler_gamma * scale
    return gumbel_r(loc, scale)


@makedist("extremevalue")
def MMextremevalue(x):
    return gumbel_r(
        x.mean() - 0.57721 * x.std() * np.sqrt(6) / np.pi, x.std() * np.sqrt(6) / np.pi
    )  # **


@makedist("lognormal")
def MMlognormal(x):
    return lognorm(x.mean(), x.std())  # **


@makedist("laplace")
def MMlaplace(x):
    return laplace(x.mean(), x.std() / np.sqrt(2))  # **


@makedist("doubleexponential")
def MMdoubleexponential(x):
    return laplace(x.mean(), x.std() / np.sqrt(2))  # **


@makedist("logistic")
def MMlogistic(x):
    return logistic(x.mean(), x.std() * np.sqrt(3) / np.pi)  # **


@makedist("loguniform")
def MMloguniform(x):
    return loguniform(mean=x.mean(), std=x.std())  # **


@makedist("norm")
def MMnormal(x):
    return norm(x.mean(), x.std())  # **


@makedist("gaussian")
def MMgaussian(x):
    return norm(x.mean(), x.std())  # **


@makedist("pareto")
def MMpareto(x):  # **
    a = x.mean()
    b = x.std()
    return pareto(a / (1 + 1 / np.sqrt(1 + a**2 / b**2)), 1 + np.sqrt(1 + a**2 / b**2))


def MMpoisson(x):
    return singleParamPattern(x, "poisson")


# test if below is correct
# def MMpoisson(x):
#     if isinstance(x, sps.CensoredData | np.ndarray | list):
#         return D.dist_from_sps(poisson(mean(x)), shape='poisson')
#     elif isinstance(x, Interval):
#         return pba.poisson(mean(x))
#     else:
#         raise TypeError('Input data type not supported')


@makedist("powerlaw")
def MMpowerfunction(x):  # **
    a = x.mean()
    b = x.std()
    return powerlaw(
        a / (1 - 1 / np.sqrt(1 + (a / b) ** 2)), np.sqrt(1 + (a / b) ** 2) - 1
    )


@makedist("t")
def MMt(x):
    return t(2 / (1 - 1 / x.std() ** 2))


@makedist("student")
def MMstudent(x):
    assert not (1 < x.std()), "Improper standard deviation for student distribution"
    return t(2 / (1 - 1 / x.std() ** 2))


@makedist("uniform")
def MMuniform(x):  # **
    a = x.mean()
    b = np.sqrt(3) * x.std()
    return uniform(a - b, a + b)


@makedist("rectangular")
def MMrectangular(x):
    return MMuniform(x)


@makedist("triangular")
def MMtriangular(x, iters=100, dives=10):  # **
    # iterative search for triangular distribution parameters using method of
    # matching moments (you solve the thing analytically! too messy without help)
    # testing code indicated with #-#
    # -#some = 10
    # -#A = runif(1,0,10)
    # -#B = A + runif(1,0,10)
    # -#C = runif(1,A,B)
    # -#x = qtriangular(runif(some), A,C,B)
    def skewness(x):
        m = x.mean()
        # std uses the population formula, may need the sample formula
        return np.sum((x - m) ** 3) / ((len(x) - 1) * x.std() ** 3)

    M = np.mean(x)
    V = np.var(x)
    S = skewness(x)
    a = aa = min(x)  # apparently double assignments work
    b = bb = max(x)
    c = cc = 3 * M - a - b
    many = iters
    s1 = np.std(x)
    for k in range(dives):
        s1 = s2 = s3 = s1 / 2
        a = np.random.normal(aa, s1, many)
        b = np.random.normal(bb, s2, many)
        c = np.random.normal(cc, s3, many)
        m = (a + b + c) / 3
        k = a**2 + b**2 + c**2 - a * b - a * c - b * c
        v = k / 18
        s = (np.sqrt(2) * (a + b - 2 * c) * (2 * a - b - c) * (a - 2 * b + c)) / (
            5 * k ** (3 / 2)
        )
        d = (M - m) ** 2 + (V - v) ** 2 + (S - s) ** 2
        i = np.argmin(d)  # which.min(d)
        aa = a[i]
        bb = b[i]
        cc = c[i]
    # -#gray(triangular(A,B,C), new = TRUE)
    # -#blue(x)
    # -#green(triangular(aa,bb,cc))
    # -#A;aa; B;bb; C;cc  # the order is min, max, mode
    print(aa, bb, cc)
    return triang(aa, cc, bb)


mom = {
    "bernoulli": MMbernoulli,
    "beta": MMbeta,
    "betabinomial": MMbetabinomial,
    "binomial": MMbinomial,
    "chisquared": MMchisquared,
    "exponential": MMexponential,
    "expon": MMexponential,
    "F": MMF,
    "f": MMF,
    "gamma": MMgamma,
    "geometric": MMgeometric,
    "gumbel": MMgumbel,
    "extremevalue": MMextremevalue,
    "lognormal": MMlognormal,
    "laplace": MMlaplace,
    "doubleexponential": MMdoubleexponential,
    "logistic": MMlogistic,
    "loguniform": MMloguniform,
    "norm": MMnormal,
    "normal": MMnormal,
    "gaussian": MMgaussian,
    "pareto": MMpareto,
    "poisson": MMpoisson,
    "powerfunction": MMpowerfunction,
    "t": MMt,
    "student": MMstudent,
    "uniform": MMuniform,
    "rectangular": MMrectangular,
    "triangular": MMtriangular,
}

###############################################################################
# * Alternative maximum likelihood estimation constructors using scipy.stats  *#
###############################################################################

# Some of these functions may support intervals in the data x.  See
# https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.rv_continuous.fit.html#scipy.stats.rv_continuous.fit
# https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.CensoredData.html#scipy.stats.CensoredData


@makedist("bernoulli")
def MLbernoulli(x):
    return bernoulli(*sps.bernoulli.fit(x))


@makedist("beta")
def MLbeta(x):
    return beta(*sps.beta.fit(x))


@makedist("betabinom")
def MLbetabinomial(x):
    return betabinom(*sps.betabinom.fit(x))


@makedist("binom")
def MLbinomial(x):
    """
    # TODO to check
    #! no fitting func for scipy discrete distributions
    """
    return binom(*sps.binom.fit(x))


@makedist("chi2")
def MLchisquared(x):
    return chi2(*sps.chi2.fit(x))


@makedist("expon")
def MLexponential(x):
    """a standalone caller for exponential distribution with interval data (not in use yet)"""
    if isinstance(x, sps.CensoredData | np.ndarray | list):
        return expon(*sps.expon.fit(x))


@makedist("f")
def MLF(x):
    return F(*sps.f.fit(x))


@makedist("gamma")
def MLgamma(x):
    return gamma(*sps.gamma.fit(x))


@makedist("gammaexpon")
def MLgammaexponential(x):
    return sps.gammaexpon(*sps.gammaexpon.fit(x))


@makedist("geom")
def MLgeometric(x):
    return geom(*sps.geom.fit(x))


@makedist("gumbel")
def MLgumbel(x):
    return gumbel_r(*sps.gumbel_r.fit(x))


@makedist("laplace")
def MLlaplace(x):
    return laplace(*sps.laplace.fit(x))


@makedist("logistic")
def MLlogistic(x):
    return logistic(*sps.logistic.fit(x))


@makedist("lognorm")
def MLlognormal(x):
    return lognorm(*sps.lognorm.fit(x))


# TODO why not use `sps.lognorm.fit(x)`` directly?
# def MLlognormal(x): return (sps.lognorm.rvs(
#     *sps.lognorm.fit(x), size=many))


@makedist("loguniform")
def MLloguniform(x):
    return loguniform(*sps.loguniform.fit(x))


@makedist("nbinom")
def MLnegativebinomial(x):
    return sps.nbinom(*sps.nbinom.fit(x))


@makedist("norm")
def MLnormal(x):
    return norm(*sps.norm.fit(x))


@makedist("pareto")
def MLpareto(x):
    return pareto(*sps.pareto.fit(x))


@makedist("poisson")
def MLpoisson(x):
    return poisson(*sps.poisson.fit(x))


@makedist("powerlaw")
def MLpowerfunction(x):
    return powerlaw(*sps.powerlaw.fit(x))


@makedist("rayleigh")
def MLrayleigh(x):
    return sps.rayleigh(*sps.rayleigh.fit(x))


@makedist("t")
def MLstudent(x):
    return t(*sps.t.fit(x))


@makedist("triang")
def MLtriangular(x):
    return triang(*sps.t.fit(x))


@makedist("uniform")
def MLuniform(x):
    return uniform(*sps.uniform.fit(x))


mle = {
    "bernoulli": MLbernoulli,
    "beta": MLbeta,
    "betabinomial": MLbetabinomial,
    "binomial": MLbinomial,
    "chisquared": MLchisquared,
    "exponential": MLexponential,
    "expon": MLexponential,
    "F": MLF,
    "f": MLF,
    "gamma": MLgamma,
    "gammaexponential": MLgammaexponential,
    "geometric": MLgeometric,
    "gumbel": MLgumbel,
    "laplace": MLlaplace,
    "logistic": MLlogistic,
    "lognormal": MLlognormal,
    "loguniform": MLloguniform,
    "negativebinomial": MLnegativebinomial,
    "norm": MLnormal,
    "normal": MLnormal,
    "gaussian": MLnormal,
    "pareto": MLpareto,
    "poisson": MLpoisson,
    "powerfunction": MLpowerfunction,
    "rayleigh": MLrayleigh,
    "student": MLstudent,
    "triangular": MLtriangular,
    "uniform": MLuniform,
}


##########################################################################
# Scott's maximum likelihood estimation constructors
##########################################################################

# Some of these functions may not support intervals in the data x           #**


def sMLbernoulli(x):
    return bernoulli(x.mean())


def sMLnormal(x):
    return norm(x.mean(), x.sd())  # **


def sMLgaussian(x):
    return MLnormal(x)


def sMLexponential(x):
    return expon(x.mean())


def sMLpoisson(x):
    return poisson(x.mean())


def sMLgeometric(x):
    return geom(1 / (1 + x.mean()))


def sMLgumbel(x):
    loc, scale = sps.gumbel_r.fit(x)
    return gumbel_r(loc, scale)


def sMLpascal(x):
    return MLgeometric(x)


def sMLuniform(x):
    return uniform(min(x), max(x))  # **


def sMLrectangular(x):
    return MLuniform(x)


def sMLpareto(x):
    return pareto(min(x), len(x) / np.sum(np.log(x) - np.log(min(x))))  # **


def sMLlaplace(x):
    return laplace(x.median(), np.sum(np.abs(x - x.median()) / len(x)))  # **


def sMLdoubleexponential(x):
    return MLlaplace(x)


def sMLlognormal2(x):  # **
    n = len(x)
    mu = np.sum(np.log(x)) / n
    # this function gives clearly poor results
    return lognormal2(mlog=mu, slog=np.sum((np.log(x) - mu) ** 2) / n)


# just uses transformation, which seems unlikely to be true, but fitdistrplus package uses it too
def sMLlognormal(x):
    return np.exp(MLnormal(np.log(x)))


def sMLloguniform(x):
    a, b, _, _ = sps.loguniform.fit(x)
    return loguniform(a, b)


def sMLweibull(x, shapeinterval=None):  # **
    if shapeinterval is None:
        shapeinterval = np.array((0.001, 500))

    def wf(k):
        return (
            np.sum(x**k * np.log(x)) / np.sum(x**k) - np.sum(np.log(x)) / len(x) - 1 / k
        )

    k = uniroot(wf, shapeinterval)
    el = np.exp(np.log(np.sum(x ^ k) / len(x)) / k)
    return sps.weibull_min.rvs(scale=el, shape=k)


def sMLgamma(data):  # **
    xbar = data.mean()
    shape = (xbar / data.sd()) ** 2  # initial estimate of shape from MoM
    logxbar = np.log(xbar)
    meanlog = np.log(data).mean()

    def f(x):
        return np.log(x) - np.digamma(x) - logxbar + meanlog

    shape = uniroot(f, shape * np.array((0.5, 1.5))).root()
    rate = shape / xbar
    return gamma(shape=shape, rate=rate)


smle = {
    "bernoulli": sMLbernoulli,
    "normal": sMLnormal,
    "gaussian": sMLgaussian,
    "exponential": sMLexponential,
    "poisson": sMLpoisson,
    "geometric": sMLgeometric,
    "gumbel": sMLgumbel,
    "pascal": sMLpascal,
    "uniform": sMLuniform,
    "rectangular": sMLrectangular,
    "pareto": sMLpareto,
    "laplace": sMLlaplace,
    "doubleexponential": sMLdoubleexponential,
    "lognormal2": sMLlognormal2,
    "lognormal": sMLlognormal,
    "loguniform": sMLloguniform,
    "weibull": sMLweibull,
    "gamma": sMLgamma,
}


##########################################################################
# Maximum entropy distribution constructors
##########################################################################


def MEminmax(min, max):
    return uniform(min, max)


# http://mathoverflow.net/questions/116667/whats-the-maximum-entropy-probability-distribution-given-bounds-a-b-and-mean, http://www.math.uconn.edu/~kconrad/blurbs/analysis/entropypost.pdf for discussion of this solution.
def MEminmaxmean(min, max, mean):
    return sawinconrad(min, mean, max)


def MEmeansd(mean, sd):
    return normal(mean, sd)


def MEminmean(min, mean):
    return min + exponential(mean - min)


# def MEmeangeomean(mean, geomean)


def MEdiscretemean(x, mu, steps=10, iterations=50):  # e.g., MEdiscretemean(1:10,2.3)
    x = np.array(x)

    def fixc(x, r):
        return 1 / np.sum(r**x)

    r = br = 1
    c = bc = fixc(x, r)
    d = bd = (mu - np.sum((c * r**x) * x)) ** 2
    for j in range(steps):
        step = 1 / (j + 1)
        for i in range(iterations):
            r = np.abs(br + (np.random.uniform() - 0.5) * step)
            c = fixc(x, r)
            d = (mu - np.sum((c * r**x) * x)) ** 2
            if d < bd:
                br = r
                bc = c
                bd = d
    w = bc * br**x
    w = w / np.sum(w)  # needed?

    z = np.array([])
    k = len(x)
    for i in range(k):
        z = np.concatenate((z, np.repeat(x[i], w[i] * many)))
    if len(z) >= many:
        z = z[0:many]
    else:
        z = np.concatenate((z, np.random.choice(x, size=many - len(z), p=w)))
    np.random.shuffle(z)  # shuffles z in place
    return z


def MEquantiles(v, p):
    if len(v) != len(p):
        stop("Inconsistent array lengths for quantiles")
    if (min(p) < 0) or (1 < max(p)):
        stop("Improper probability for quantiles")  # ensure 0 <= p <= 1
    if not (min(p) == 0 and max(p) == 1):
        stop("Probabilities must start at zero and go to one for quantiles")
    if np.any(np.diff(p) < 0):
        # ensure montone probabilities
        stop("Probabilities must increase for quantiles")
    if np.any(np.diff(v) < 0):
        stop("Quantiles values must increase")  # ensure montone quantiles
    x = np.repeat(np.inf, many)
    r = np.random.uniform(size=many)
    # np.where is Python's version of R's ifelse function
    for i in range(len(p) - 1):
        x = np.where(
            (p[i] <= r) & (r < p[i + 1]),
            v[i] + (r - p[i]) * (v[i + 1] - v[i]) / (p[i + 1] - p[i]),
            x,
        )
    return x


def MEdiscreteminmax(min, max):
    return np.minimum(np.trunc(uniform(min, max + 1)), max)


def MEmeanvar(mean, var):
    return MEmeansd(mean, np.sqrt(var))


def MEminmaxmeansd(min, max, mean, sd):
    return beta1((mean - min) / (max - min), sd / (max - min)) * (max - min) + min


def MEmmms(min, max, mean, sd):
    return beta1((mean - min) / (max - min), sd / (max - min)) * (max - min) + min


def MEminmaxmeanvar(min, max, mean, var):
    return MEminmaxmeansd(min, max, mean, np.sqrt(var))


###############################################################################
# Miscellaneous: PERT, Fermi methods, mean-normal range, KS, EDF, Antweiler
###############################################################################


# https://wernerantweiler.ca/blog.php?item=2019-06-05       #**
def antweiler(x):
    return triang(min=min(x), mode=3 * np.mean(x) - max(x) - min(x), max=max(x))


def betapert(min, max, mode):  # N.B.  Not in numerical order!
    mu = (min + max + 4 * mode) / 6
    if abs(mode - mu) < 1e-8:
        alpha1 = alpha2 = 3
    else:
        alpha1 = (mu - min) * (2 * mode - min - max) / ((mode - mu) * (max - min))
        alpha2 = alpha1 * (max - mu) / (mu - min)
    return min + (max - min) * beta(alpha1, alpha2)


def mnr(n, many=10000):
    xL = xU = np.random.normal(size=many)
    for i in range(n - 1):
        xx = np.random.normal(size=many)
        xL = np.minimum(xL, xx)
        xU = np.maximum(xU, xx)
    return np.mean(xU - xL)


def fermilnorm(x1, x2, n=None, pr=0.9):
    gm = np.sqrt(x1 * x2)
    if n is None:
        gsd = np.sqrt(x2 / x1) ** (1 / scipy.stats.norm.ppf(pr))  # qnorm(pr)
    else:
        gsd = np.exp((np.log(x2) - np.log(x1)) / mnr(n))
    return np.log((gm, gsd))


def ferminorm(x1, x2, n=None, pr=0.9):
    m = (x1 + x2) / 2
    if n is None:
        s = (x2 - x1) / (2 * scipy.stats.norm.ppf(pr))  # qnorm(pr)
    else:
        s = (x2 - x1) / mnr(n)
    return np.array((m, s))


def approxksD95(n):
    from scipy.interpolate import CubicSpline

    # approximations for the critical level for Kolmogorov-Smirnov statistic D,
    # for confidence level 0.95. Taken from Bickel & Doksum, table IX, p.483
    # and Lienert G.A.(1975) who attributes to Miller,L.H.(1956), JASA
    if n > 80:
        # Bickel&Doksum, table IX,p.483
        return 1.358 / (np.sqrt(n) + 0.12 + 0.11 / np.sqrt(n))
    else:
        x = np.array(
            (1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 15, 20, 30, 40, 50, 60, 70, 80)
        )  # from Lienert
        y = np.array(
            (
                0.975,
                0.84189,
                0.70760,
                0.62394,
                0.56328,  # 1:5
                0.51926,
                0.48342,
                0.45427,
                0.43001,
                0.40925,  # 6:10
                0.33760,
                0.29408,
                0.24170,
                0.21012,  # 15,20,30,40
                0.18841,
                0.17231,
                0.15975,
                0.14960,
            )
        )  # 50,60,70,80
        f = CubicSpline(x, y, bc_type="natural")
        return f(n)


def ks(x, conf=0.95, min=None, max=None):
    if conf != 0.95:
        stop("Cannot currently handle confidence levels other than 95%")
    h = histogram(x)
    mn = np.min(x)
    mx = np.max(x)
    if min is None:
        min = mn - (mx - mn) / 2
    if max is None:
        max = mx + (mx - mn) / 2
    lots = int(approxksD95(len(x)) * many)
    Lfermi = np.concatenate((lots * [min], h))
    Rfermi = np.concatenate((h, lots * [max]))
    Lfermi.sort()
    Rfermi[::-1].sort()
    # should prolly shuffle
    return np.concatenate((Lfermi[0:many], Rfermi[0:many]))


def ferminormconfband(x1, x2, n, pr=0.9, conf=0.95, bOt=0.001, tOp=0.999):
    if conf != 0.95:
        stop("Cannot handle confidence levels other than 95%")
    m, s = ferminorm(x1, x2, n, pr)
    lots = int(approxksD95(n) * many)
    BOT = scipy.stats.norm.ppf(bOt, m, s)
    TOP = scipy.stats.norm.ppf(tOp, m, s)
    Lfermi = np.concatenate((lots * [BOT], scipy.stats.norm.rvs(m, s, size=many)))
    Rfermi = np.concatenate((lots * [TOP], scipy.stats.norm.rvs(m, s, size=many)))
    Lfermi.sort()
    Rfermi[::-1].sort()
    # should prolly shuffle
    return np.concatenate((Lfermi[0:many], Rfermi[0:many]))


def fermilnormconfband(x1, x2, n, pr=0.9, conf=0.95, bOt=0.001, tOp=0.999):
    if conf != 0.95:
        stop("Cannot handle confidence levels other than 95%")
    mlog, slog = fermilnorm(x1, x2, n, pr)
    d = lognormal2(mlog, slog)
    lots = int(approxksD95(n) * many)
    BOT = scipy.stats.lognorm.ppf(bOt, s=slog, scale=np.exp(mlog))
    TOP = scipy.stats.lognorm.ppf(tOp, s=slog, scale=np.exp(mlog))
    Lfermi = np.concatenate((lots * [BOT], d))
    Rfermi = np.concatenate((lots * [TOP], d))
    Lfermi.sort()
    Rfermi[::-1].sort()
    # should prolly shuffle
    return np.concatenate((Lfermi[0:many], Rfermi[0:many]))


# * --------------- moments shape distribution constructors --------------- *#


def mmmshape(mean, var, shape):
    """return a distributional object based on moment and shape information"""
    if shape == "beta":
        return mm_beta(mean, var)
    elif shape == "normal":
        return mm_normal(mean, var)
    elif shape == "gamma":
        return mm_gamma(mean, var)
    else:
        raise ValueError(f"Shape '{shape}' not supported for moment matching.")


def mm_beta(mean_sample, var_sample):
    """return distributional parameters for beta distribution"""
    common = mean_sample * (1 - mean_sample) / var_sample - 1
    alpha = mean_sample * common
    beta_param = (1 - mean_sample) * common

    return alpha, beta_param


def mm_normal(mean_sample, var_sample):
    """return distributional parameters for normal distribution"""
    return mean_sample, np.sqrt(var_sample)


def mm_gamma(mean_sample, var_sample):
    """return distributional parameters for gamma distribution"""
    alpha = mean_sample**2 / var_sample
    theta = var_sample / mean_sample

    return alpha, theta


# TODO: add more distributions in the future



from __future__ import annotations
from typing import TYPE_CHECKING
from ...pba.intervals.number import Interval

import numpy as np
from rich.progress import track
from typing import Callable
from .cartesian_product import cartesian
from ..utils import Propagation_results


# TODO it currently takes 2D numpy array as input
def endpoints_method(
    x: Interval | np.ndarray,
    f: Callable,
    save_raw_data=False,
    results: Propagation_results = None,
) -> Propagation_results:
    """
        Performs uncertainty propagation using the endpoints or vertex method.

    args:
        x (np.ndarray): A 2D NumPy array where each row represents an input variable and
          the two columns define its lower and upper bounds (interval).
        f (Callable): A callable function that takes a 1D NumPy array of input values and
          returns the corresponding output(s).
        results (Propagation_results): The class to use for storing results (defaults to Propagation_results).
        save_raw_data (boolean, optional): Acts as a switch to enable or disable the storage of raw input data when a function (f)
          is not provided.
          - 'False': Returns an error that no function is provided.
          - 'True': Returns the full arrays of unique input combinations.

    notes:
        The function assumes that the intervals in `x` represent uncertainties and aims to provide conservative bounds on the output
        uncertainty.
        If the `f` function returns multiple outputs, the `bounds` array will be 2-dimensional.

    return:
        Returns `Propagation_results` object(s) containing:
            - 'un': UncertainNumber object(s) to characterise the interval(s) of the output(s).
            - 'raw_data' (dict): Dictionary containing raw data shared across output(s):
                    - 'x' (np.ndarray): Input values.
                    - 'f' (np.ndarray): Output values.
                    - 'min' (np.ndarray): Array of dictionaries, one for each output,
                              containing 'f' for the minimum of that output.
                    - 'max' (np.ndarray): Array of dictionaries, one for each output,
                              containing 'f' for the maximum of that output.
                    - 'bounds' (np.ndarray): 2D array of lower and upper bounds for each output.

    raises:
        ValueError if no function is provided and save_raw_data is 'no'.

    example:
        # Example usage with different parameters for minimization and maximization
        f = lambda x: x[0] + x[1] + x[2]  # Example function

        # Determine input parameters for function and method
        x_bounds = np.array([[1, 2], [3, 4], [5, 6]])
        y = endpoints_method(x_bounds, f)

    """
    if isinstance(x, Interval):
        x = x.to_numpy()

    if results is None:
        results = Propagation_results()  # Create an instance of Propagation_results

    # Create a sequence of values for each interval based on the number of divisions provided
    # The divisions may be the same for all intervals or they can vary.
    print(
        f"Total number of input combinations for the endpoint method: {2**x.shape[0]}"
    )

    # create an array with the unique combinations of all intervals
    X = cartesian(*x)

    # propagates the epistemic uncertainty through subinterval reconstitution
    if f is not None:
        all_output = np.array(
            [f(xi) for xi in track(X, description="Function evaluations")]
        )

        try:
            num_outputs = len(all_output[0])
        except TypeError:
            num_outputs = 1  # If f returns a single value

        # Reshape all_output to a 2D array (Corrected)
        all_output = np.array(all_output).reshape(-1, num_outputs)

        if all_output.shape[1] == 1:  # Single output
            results.raw_data["bounds"] = np.array(
                [np.min(all_output, axis=0)[0], np.max(all_output, axis=0)[0]]
            )
        else:  # Multiple outputs
            bounds = np.empty((all_output.shape[1], 2))
            for i in range(all_output.shape[1]):
                bounds[i, :] = np.array(
                    [np.min(all_output[:, i], axis=0), np.max(all_output[:, i], axis=0)]
                )
            results.raw_data["bounds"] = bounds

        for i in range(num_outputs):  # Iterate over outputs
            min_indices = np.where(
                all_output[:, i] == np.min(all_output[:, i], axis=0)
            )[0]
            max_indices = np.where(
                all_output[:, i] == np.max(all_output[:, i], axis=0)
            )[0]

            # Convert to 2D arrays and append
            results.raw_data["min"] = np.append(
                results.raw_data["min"],
                {"x": X[min_indices], "f": np.min(all_output[:, i], axis=0)},
            )
            results.raw_data["max"] = np.append(
                results.raw_data["max"],
                {"x": X[max_indices], "f": np.max(all_output[:, i], axis=0)},
            )

        results.raw_data["f"] = all_output
        results.raw_data["x"] = X

    elif save_raw_data:  # If f is None and save_raw_data is 'yes'
        # Store X in raw_data['x'] even if f is None
        results.add_raw_data(x=X)

    else:
        raise Exception(
            "No function is provided. Select save_raw_data = 'yes' to save the input combinations"
        )

    return results



import numpy as np
from typing import Callable, Union
import tqdm
from pyuncertainnumber.propagation.epistemic_uncertainty.extreme_point_func import extreme_pointX
from pyuncertainnumber.propagation.epistemic_uncertainty.extremepoints import extremepoints_method
from pyuncertainnumber.propagation.utils import Propagation_results, condense_bounds


def imp(X):
    """Imposition of intervals."""

    return np.array([np.max(X[:, 0]), np.min(X[:, 1])])


def first_order_propagation_method(x: list, f: Callable = None,
                                   results: Propagation_results = None,
                                   # method = 'extremepoints',
                                   n_disc: Union[int, np.ndarray] = 10,
                                   condensation: Union[float,
                                                       np.ndarray] = None,
                                   tOp: Union[float, np.ndarray] = 0.999,
                                   bOt: Union[float, np.ndarray] = 0.001,
                                   save_raw_data='no') -> Propagation_results:  # Specify return type
    """
    args:
        - x (list): A list of `UncertainNumber` objects representing the uncertain inputs.
        - f (Callable): The function to evaluate.
        - results (Propagation_results, optional): An object to store propagation results.
                                    Defaults to None, in which case a new
                                    `Propagation_results` object is created.
        - n_disc (Union[int, np.ndarray], optional): The number of discretization points 
                                    for each uncertain input. If an integer 
                                    is provided, it is used for all inputs. 
                                    If a NumPy array is provided, each element 
                                    specifies the number of discretization 
                                    points for the corresponding input. 
                                    Defaults to 10.
        - condensation (Union[float, np.ndarray], optional): A parameter or array of parameters 
                                    to control the condensation of the output p-boxes. 
                                    Defaults to None.
        - tOp (Union[float, np.ndarray], optional): Upper threshold or array of thresholds for 
                                    discretization. 
                                    Defaults to 0.999.
        - bOt (Union[float, np.ndarray], optional): Lower threshold or array of thresholds for 
                                    discretization. 
                                    Defaults to 0.001.
        - save_raw_data (str, optional): Whether to save raw data ('yes' or 'no'). 
                                   Defaults to 'no'.

    signature:
        first_order_propagation_method(x: list, f: Callable, results: Propagation_results = None, ...) -> Propagation_results

    notes:
        - Performs first-order uncertainty propagation for mixed uncertain numbers.
        - The function handles different types of uncertain numbers (distributions 
          and p-boxes for this version) and discretizes them with the same number of n_disc.
        - It uses the `extremepoints` to determine the signs of the partial 
          derivatives of the function.
        - The output p-boxes are constructed by imposing the results from individual 
          input discretizations.
        - The `condensation` parameter can be used to reduce the number of intervals in 
          the output p-boxes.

    returns:
        Propagation_results: A `Propagation_results` object containing the results of the 
                          uncertainty propagation. The results include p-boxes representing 
                          the output uncertainty.

    example:
        from pyuncertainnumber import UncertainNumber

        def Fun(x):

            input1= x[0]
            input2=x[1]
            input3=x[2]
            input4=x[3]
            input5=x[4]

            output1 = input1 + input2 + input3 + input4 + input5
            output2 = input1 * input2 * input3 * input4 * input5

            return np.array([output1, output2])

        means = np.array([1, 2, 3, 4, 5])
        stds = np.array([0.1, 0.2, 0.3, 0.4, 0.5])

        x = [
            UncertainNumber(essence = 'distribution', distribution_parameters= ["gaussian",[means[0], stds[0]]]),

        from pyuncertainnumber import UncertainNumber

        def Fun(x):

            input1= x[0]
            input2=x[1]
            input3=x[2]
            input4=x[3]
            input5=x[4]

            output1 = input1 + input2 + input3 + input4 + input5
            output2 = input1 * input2 * input3 * input4 * input5

            return np.array([output1, output2])

        means = np.array([1, 2, 3, 4, 5])
        stds = np.array([0.1, 0.2, 0.3, 0.4, 0.5])

        x = [
            UncertainNumber(essence = 'distribution', distribution_parameters= ["gaussian",[means[0], stds[0]]]),

            UncertainNumber(essence = 'distribution', distribution_parameters= ["gaussian",[means[1], stds[1]]]),
            UncertainNumber(essence = 'distribution', distribution_parameters= ["gaussian",[means[2], stds[2]]]),
            UncertainNumber(essence = 'distribution', distribution_parameters= ["gaussian",[means[3], stds[3]]]),
            UncertainNumber(essence = 'distribution', distribution_parameters= ["gaussian",[means[4], stds[4]]]),
            ]

        results = first_order_propagation_method(x=x, f=Fun, n_disc= 5)
    """
    d = len(x)  # dimension of uncertain numbers
    results = Propagation_results()
    xl = []
    xr = []
    ranges = np.zeros((2, d))
    n_slices = np.zeros((d), dtype=int)
    u_list = []  # List to store 'u' for each uncertain number.

    for i, un in enumerate(x):
        print(f"Processing variable {i + 1} with essence: {un.essence}")

        if un.essence != "interval":

            distribution_family = un.distribution_parameters[0]

            # perform outward directed discretisation
            if isinstance(n_disc, int):
                nd = n_disc + 1
            else:
                nd = n_disc[i] + 1  # Use the i-th element of n_disc array
            n_slices[i] = nd

            if isinstance(tOp, np.ndarray):
                top = tOp[i]
            else:
                top = tOp
            if isinstance(bOt, np.ndarray):
                bot = bOt[i]
            else:
                bot = bOt

            if distribution_family == 'triang':
                u = np.linspace(0, 1, nd)
            else:
                u = np.linspace(bot, top, nd)  # Use bOt and tOp here

        else:  # un.essence == "interval"
            u = np.array([0.0, 1.0])  # Or adjust as needed for intervals
            n_slices[i] = 2

        u_list.append(u)  # Add 'u' to the list

        # Generate discrete p-boxes
        temp_xl = []  # Temporary list to hold xl values for the current variable
        temp_xr = []  # Temporary list to hold xr values for the current variable

        match un.essence:
            case "distribution":
                # Calculate xl and xr for distributions (adjust as needed)
                # Assuming un.ppf(u) returns a list or array
                temp_xl = un.ppf(u_list[i][:-1]).tolist()
                # Adjust based on your distribution
                temp_xr = un.ppf(u_list[i][1:]).tolist()
                ranges[:, i] = np.array([temp_xl[0], temp_xr[-1]])

            case "interval":
                # Repeat lower bound for all quantiles
                temp_xl = np.array([un.bounds[0]]).tolist()
                # Repeat upper bound for all quantiles
                temp_xr = np.array([un.bounds[1]]).tolist()
                ranges[:, i] = np.array([un.bounds])

            case "pbox":
                # Assuming un.ppf(u) returns a list of lists
                temp_xl = un.ppf(u_list[i][:-1])[0].tolist()
                temp_xr = un.ppf(u_list[i][1:])[1].tolist()
                ranges[:, i] = np.array([temp_xl[0], temp_xr[-1]])
            case _:
                raise ValueError(f"Unsupported uncertainty type: {un.essence}")

        xl.append(temp_xl)  # Add the temporary list to xl
        xr.append(temp_xr)  # Add the temporary list to xr
    if f is not None:
        # Determine the positive or negative signs for each input
        res = extremepoints_method(ranges.T, f)

        # Determine the number of outputs from the first evaluation
        try:
            num_outputs = res.raw_data['sign_x'].shape[0]
        except TypeError:
            num_outputs = 1  # If f returns a single value

        all_output_list = []
        for _ in range(num_outputs):
            output_for_current_output = []
            for input in range(d):
                slices_for_input = []
                for _ in range(n_slices[input]-1):  # Use n_slices[input] here
                    slices_for_input.append([None, None])
                output_for_current_output.append(slices_for_input)
            all_output_list.append(output_for_current_output)

        inpsList = np.zeros((0, d))
        evalsList = np.zeros((0, num_outputs))

        for input in range(d):  # Iterate over input variables first
            X = [ranges[:, k].tolist() for k in range(d)]
            temp_X = X.copy()
            Xsings = np.empty((n_slices[input], 2, d))
            current_index = 0

            for slice in tqdm.tqdm(range(n_slices[input] - 1), desc=f"Processing input {input+1}"):
                temp_X[input] = []
                temp_X[input].extend(
                    np.array([xl[input][slice], xr[input][slice]]).tolist())
                rang = np.array([temp_X[i] for i in range(d)], dtype=object)
                Xsings[slice, :, :] = extreme_pointX(
                    rang, res.raw_data['sign_x'])  # Use the entire sign_x array
                current_index += 1

                for k in range(Xsings.shape[1]):
                    c = Xsings[slice, k, :]
                    im = np.where((inpsList == c).all(axis=1))[0]
                    if not im.size:
                        output = f(c)

                        # Store each output in a separate sublist
                        for out in range(num_outputs):
                            all_output_list[out][input][slice][k] = output[out]

                        inpsList = np.vstack([inpsList, c])
                        evalsList = np.vstack([evalsList, np.array(output)])
                    else:
                        for out in range(num_outputs):
                            all_output_list[out][input][slice][k] = evalsList[im[0]][out]

                current_index = 0  # Reset for the next input

        # Reshape all_output based on the actual number of elements per output
        all_output = np.array(all_output_list, dtype=object)

        all_output = np.reshape(
            all_output, (num_outputs, d, -1, 2))  # Reshape to 4D

        # Calculate min and max for each output and input variable
        min_values = np.min(all_output, axis=3)
        max_values = np.max(all_output, axis=3)

        # Initialize bounds_input
        bounds_input = np.empty((num_outputs, d, n_disc, 2))

        for i in range(num_outputs):
            for j in range(d):  # Iterate over input variables
                # Merge min_values and max_values into bounds_input
                for k in range(n_disc):
                    bounds_input[i, j, k, 0] = min_values[i, j, k]
                    bounds_input[i, j, k, 1] = max_values[i, j, k]

                # Sort bounds_input along the last axis (k)
                bounds_input[i, j, :, :] = np.sort(
                    bounds_input[i, j, :, :], axis=-1)

        bounds = np.empty((num_outputs, 2, n_disc))  # Initialize bounds
        lower_bound = np.zeros((num_outputs, n_disc))  # Initialize lower_bound
        upper_bound = np.zeros((num_outputs, n_disc))  # Initialize upper_bound

        for i in range(num_outputs):
            temp_bounds = []  # Temporary list for bounds

            for k in range(n_disc):  # Iterate over input variables
                # Impose the p-boxes for each input variable and store in temporary list
                temp_bounds.append(imp(bounds_input[i, :, k, :]))

            # Impose the p-boxes across all input variables for the current output
            bounds[i, :, :] = np.array(temp_bounds).T

            # Extract lower_bound and upper_bound from bounds
            lower_bound[i, :] = bounds[i, 1, :]
            upper_bound[i, :] = bounds[i, 0, :]

        if condensation is not None:
            bounds = condense_bounds(bounds, condensation)

        results.raw_data['bounds'] = bounds
        results.raw_data['min'] = np.array([{'f': lower_bound[i, :]} for i in range(
            num_outputs)])  # Initialize as a NumPy array
        results.raw_data['max'] = np.array([{'f': upper_bound[i, :]} for i in range(
            num_outputs)])  # Initialize as a NumPy array

        if save_raw_data == 'yes':
            print('No raw data provided for this method!')
            # results.add_raw_data(f= all_output, x= x_combinations)

    elif save_raw_data == 'yes':  # If f is None and save_raw_data is 'yes'
        # results.add_raw_data(f= None, x= x_combinations)
        print('No raw data provided for this method!')

    else:
        print("No function is provided. Please provide a function!")

    return results

# from pyuncertainnumber import UncertainNumber

# def myFunctionWithTwoOutputs(x):
#     """
#     Example function with two outputs.
#     Replace this with your actual function logic.
#     """
#     input1= x[0]
#     input2=x[1]
#     input3=x[2]
#     input4=x[3]
#     input5=x[4]
#     output1 = input1 + input2 + input3 + input4 + input5
#     output2 = input1 * input2 * input3 * input4 * input5
#     return output1, output2

# means = np.array([1, 2, 3, 4, 5])
# stds = np.array([0.1, 0.2, 0.3, 0.4, 0.5])
# x = np.array([1, 2, 3, 4, 5])
# a = myFunctionWithTwoOutputs(x)
# print(a[0])

# n_disc = 1000  # Number of discretizations

# x = [
#         UncertainNumber(essence='distribution', distribution_parameters=["gaussian",[means[0], stds[0]]]),
#         UncertainNumber(essence = 'distribution', distribution_parameters= ["gaussian",[means[1], stds[1]]]),
#         UncertainNumber(essence = 'distribution', distribution_parameters= ["gaussian",[means[2], stds[2]]]),
#         UncertainNumber(essence = 'distribution', distribution_parameters= ["gaussian",[means[3], stds[3]]]),
#         UncertainNumber(essence = 'distribution', distribution_parameters= ["gaussian",[means[4], stds[4]]]),
#     ]

# results = first_order_propagation_method(x=x, f=myFunctionWithTwoOutputs, n_disc=n_disc)

# print(results)

# import matplotlib.pyplot as plt

# def plotPbox(xL, xR, p=None):
#     """
#     Plots a p-box (probability box) using matplotlib.

#     Args:
#         xL (np.ndarray): A 1D NumPy array of lower bounds.
#         xR (np.ndarray): A 1D NumPy array of upper bounds.
#         p (np.ndarray, optional): A 1D NumPy array of probabilities corresponding to the intervals.
#                                    Defaults to None, which generates equally spaced probabilities.
#         color (str, optional): The color of the plot. Defaults to 'k' (black).
#     """
#     xL = np.squeeze(xL)  # Ensure xL is a 1D array
#     xR = np.squeeze(xR)  # Ensure xR is a 1D array

#     if p is None:
#         p = np.linspace(0, 1, len(xL))  # p should have one more element than xL/xR

#     if p.min() > 0:
#         p = np.concatenate(([0], p))
#         xL = np.concatenate(([xL[0]], xL))
#         xR = np.concatenate(([xR[0]], xR))

#     if p.max() < 1:
#         p = np.concatenate((p, [1]))
#         xR = np.concatenate((xR, [xR[-1]]))
#         xL = np.concatenate((xL, [xL[-1]]))

#     colors = 'black'
#     # Highlight the points (xL, p)
#     plt.scatter(xL, p, color=colors, marker='o', edgecolors='black', zorder=3)

#     # Highlight the points (xR, p)
#     plt.scatter(xR, p, color=colors, marker='o', edgecolors='black', zorder=3)


#     plt.fill_betweenx(p, xL, xR, color=colors, alpha=0.5)
#     plt.plot( [xL[0], xR[0]], [0, 0],color=colors, linewidth=3)
#     plt.plot([xL[-1], xR[-1]],[1, 1],  color=colors, linewidth=3)
#     plt.show()

# plotPbox(results.raw_data['min'][1]['f'], results.raw_data['max'][1]['f'], p=None)
# plt.show()

# plotPbox(results.raw_data['min'][0]['f'], results.raw_data['max'][0]['f'], p=None)
# plt.show()



import numpy as np


def extreme_pointX(ranges, signX):
    """ Calculates the extreme points of a set of ranges based on signs.

    args:
        ranges: A NumPy array of shape (d, 2) representing the ranges 
                 (each row is a variable, each column is a bound).
        signX: A NumPy array of shape (1, d) representing the signs.

    returns:
        A NumPy array of shape (2, d) representing the extreme points.
    """
    d = len(ranges)  # Get the number of dimensions
    pts = np.tile(signX, (2, 1))  # Repeat signX twice vertically

    Xsign = np.zeros((2, d))
    for j in range(d):  # Iterate over dimensions
        if pts[0, j] > 0:  # Check sign for the first row (minimum)
            Xsign[0, j] = ranges[j][0]  # Take the first element (lower bound)
            Xsign[1, j] = ranges[j][-1]  # Take the last element (upper bound)
        else:
            Xsign[0, j] = ranges[j][-1]  # Take the last element (upper bound)
            Xsign[1, j] = ranges[j][0]  # Take the first element (lower bound)

    return Xsign



from .intervals.number import Interval as I
from .pbox_parametric import *
from .pbox_free import *
from .dss import DempsterShafer
from .distributions import Distribution, JointDistribution, ECDF
from .dependency import Dependency



import numpy as np
from typing import Callable, Union
import tqdm
from pyuncertainnumber.propagation.epistemic_uncertainty.extreme_point_func import extreme_pointX
from pyuncertainnumber.propagation.epistemic_uncertainty.extremepoints import extremepoints_method
from pyuncertainnumber.propagation.utils import Propagation_results, condense_bounds

def imp(X):
    """Imposition of intervals."""

    return np.array([np.max(X[:, 0]), np.min(X[:, 1])])

def varied_discretisation_propagation_method(x: list, f:Callable = None,  
                                results: Propagation_results = None, 
                                method = 'extremepoints',
                                n_disc: Union[int, np.ndarray] = 10, 
                                tOp: Union[float, np.ndarray] = 0.999,
                                bOt: Union[float, np.ndarray] = 0.001,
                                condensation: Union[float, np.ndarray] = None, 
                                save_raw_data= 'no')-> Propagation_results:  # Specify return type

    """ Performs uncertainty propagation for mixed uncertain numbers. The function handles different types of uncertain 
        numbers (distributions and p-boxes for this version) and discretizes them with the same number of n_disc.To ensure 
        conservative results, the function employs an outward-directed discretization approach when discretizing probability
        distributions and pboxes.For distributions that extend to infinity (e.g., normal distribution), the discretization 
        process incorporates cut-off points defined by the tOp (upper) and bOt (lower) parameters to bound the distribution.
        The function analyses the effect of each input individually assuming all other are reduced to intervals. It uses the
        `extremepoints` to determine the signs of the partial derivatives of the function. For each input, the function 
        constructs a pbox output.  It then combines these individual p-boxes by finding the overlapping region of uncertainty
        that is common to all of them.  This overlapping region represents the overall uncertainty in the output(s). The 
        `condensation` parameter can be used to reduce the number of intervals in the output p-boxes.        

    args:
        x (list): A list of `UncertainNumber` objects representing the uncertain inputs.
        f (Callable): The function to evaluate.
        results (Propagation_results, optional): An object to store propagation results.
                                    Defaults to None, in which case a new
                                    `Propagation_results` object is created.
        method (str, optional): The method to use.
                            - 'extremepoints', 'varied_discretisation_extremepoints': Only method 
                                supported. 
        n_disc (Union[int, np.ndarray], optional): The number of discretization points 
                                    for each uncertain input. If an integer 
                                    is provided, it is used for all inputs. 
                                    If a NumPy array is provided, each element 
                                    specifies the number of discretization 
                                    points for the corresponding input. 
                                    Defaults to 10.
        tOp (Union[float, np.ndarray], optional): Upper threshold or array of thresholds for 
                                    discretization. 
                                    Defaults to 0.999.
        bOt (Union[float, np.ndarray], optional): Lower threshold or array of thresholds for 
                                    discretization. 
                                    Defaults to 0.001.
        condensation (Union[float, np.ndarray], optional): A parameter or array of parameters 
                                    to control the condensation of the output p-boxes. 
                                    Defaults to None.
        save_raw_data: Acts as a switch to enable or disable the storage of raw input data when a function (f) 
                                    is not provided.
                                    - 'no': Returns an error that no function is provided.
                                    - 'yes': Returns the full arrays of unique input combinations.

    signature:
        varied_discretisation_propagation_method(x: list, f: Callable,  ...) -> Propagation_results
   
    notes:
        It is more efficient and more conserative than the focused_discretisation_propagation_method.
    
    returns:
        Returns `Propagation_results` object(s) containing:
            - 'un': UncertainNumber object(s) to characterise the empirical pbox of the output(s).
            - 'raw_data' (dict): Dictionary containing raw data shared across output(s):
                    - 'x' (np.ndarray): Input values.
                    - 'f' (np.ndarray): Output values.
                    - 'min' (np.ndarray): Array of dictionaries, one for each output,
                              containing 'f' for the minimum of that output.
                    - 'max' (np.ndarray): Array of dictionaries, one for each output,
                              containing 'f' for the maximum of that output.
                    - 'bounds' (np.ndarray): 2D array of lower and upper bounds for each output.

    example:
        from PyUncertaiNnumber import UncertainNumber

        def Fun(x):

            input1= x[0]
            input2=x[1]
            input3=x[2]
            input4=x[3]
            input5=x[4]

            output1 = input1 + input2 + input3 + input4 + input5
            output2 = input1 * input2 * input3 * input4 * input5

            return np.array([output1, output2])

        means = np.array([1, 2, 3, 4, 5])
        stds = np.array([0.1, 0.2, 0.3, 0.4, 0.5])

        x = [
            UncertainNumber(essence = 'distribution', distribution_parameters= ["gaussian",[means[0], stds[0]]]),

            UncertainNumber(essence = 'distribution', distribution_parameters= ["gaussian",[means[1], stds[1]]]),
            UncertainNumber(essence = 'distribution', distribution_parameters= ["gaussian",[means[2], stds[2]]]),
            UncertainNumber(essence = 'distribution', distribution_parameters= ["gaussian",[means[3], stds[3]]]),
            UncertainNumber(essence = 'distribution', distribution_parameters= ["gaussian",[means[4], stds[4]]])
            ]
        
        results = varied_discretisation_propagation_method(x=x, f=Fun, n_disc= 5)
    """
    d = len(x)  # dimension of uncertain numbers
    results = Propagation_results()
    xl = []
    xr = []
    ranges = np.zeros((2, d))
    n_slices = np.zeros((d), dtype=int)
    u_list = []  # List to store 'u' for each uncertain number.

    for i, un in enumerate(x):
        print(f"Processing variable {i + 1} with essence: {un.essence}")

        if un.essence != "interval":

            distribution_family = un.distribution_parameters[0]

            # perform outward directed discretisation
            if isinstance(n_disc, int):
                nd = n_disc + 1
            else:
                nd = n_disc[i] + 1  # Use the i-th element of n_disc array
            n_slices[i] = nd

            if isinstance(tOp, np.ndarray):
                top = tOp[i]
            else:
                top = tOp
            if isinstance(bOt, np.ndarray):
                bot = bOt[i]
            else:
                bot = bOt

            if distribution_family == 'triang':
                u = np.linspace(0, 1, nd)
            else:
                u = np.linspace(bot, top, nd)  # Use bOt and tOp here

        else:  # un.essence == "interval"
            u = np.array([0.0, 1.0])  # Or adjust as needed for intervals
            n_slices[i] = 2

        u_list.append(u)  # Add 'u' to the list

        # Generate discrete p-boxes
        temp_xl = []  # Temporary list to hold xl values for the current variable
        temp_xr = []  # Temporary list to hold xr values for the current variable

        match un.essence:
            case "distribution":
                # Calculate xl and xr for distributions (adjust as needed)
                # Assuming un.ppf(u) returns a list or array
                temp_xl = un.ppf(u_list[i][:-1]).tolist()
                # Adjust based on your distribution
                temp_xr = un.ppf(u_list[i][1:]).tolist()
                ranges[:, i] = np.array([temp_xl[0], temp_xr[-1]])

            case "interval":
                # Repeat lower bound for all quantiles
                temp_xl = np.array([un.bounds[0]]).tolist()
                # Repeat upper bound for all quantiles
                temp_xr = np.array([un.bounds[1]]).tolist()
                ranges[:, i] = np.array([un.bounds])

            case "pbox":
                # Assuming un.ppf(u) returns a list of lists
                temp_xl = un.ppf(u_list[i][:-1])[0].tolist()
                temp_xr = un.ppf(u_list[i][1:])[1].tolist()
                ranges[:, i] = np.array([temp_xl[0], temp_xr[-1]])
            case _:
                raise ValueError(f"Unsupported uncertainty type: {un.essence}")

        xl.append(temp_xl)  # Add the temporary list to xl
        xr.append(temp_xr)  # Add the temporary list to xr
    
    match method:
        case "extremepoints" | "varied_discretisation_extremepoints":
            if f is not None:
                # Determine the positive or negative signs for each input
                res = extremepoints_method(ranges.T, f)

                # Determine the number of outputs from the first evaluation
                try:
                    num_outputs = res.raw_data['part_deriv_sign'].shape[0]
                except TypeError:
                    num_outputs = 1  # If f returns a single value

                all_output_list = []
                for _ in range(num_outputs):
                    output_for_current_output = []
                    for input in range(d):
                        slices_for_input = []
                        for _ in range(n_slices[input]-1):  # Use n_slices[input] here
                            slices_for_input.append([None, None])
                        output_for_current_output.append(slices_for_input)
                    all_output_list.append(output_for_current_output)

                inpsList = np.zeros((0, d))
                evalsList = np.zeros((0, num_outputs))
                num_evaluations = 0  # Reset for the next input

                for input in range(d):  # Iterate over input variables first
                    X = [ranges[:, k].tolist() for k in range(d)]
                    temp_X = X.copy()
                    Xsings = np.empty((n_slices[input], 2, d))

                    for slice in tqdm.tqdm(range(n_slices[input] - 1), desc=f"Processing input {input+1}"):
                        temp_X[input] = []
                        temp_X[input].extend(
                            np.array([xl[input][slice], xr[input][slice]]).tolist())
                        rang = np.array([temp_X[i] for i in range(d)], dtype=object)
                        Xsings[slice, :, :] = extreme_pointX(rang, res.raw_data['part_deriv_sign'])  # Use the entire part_deriv_sign array
                        
                        for k in range(Xsings.shape[1]):
                            c = Xsings[slice, k, :]
                            im = np.where((inpsList == c).all(axis=1))[0]
                            if not im.size:
                                output = f(c)
                                num_evaluations += 1

                                # Store each output in a separate sublist
                                for out in range(num_outputs):
                                    all_output_list[out][input][slice][k] = output[out]

                                inpsList = np.vstack([inpsList, c])
                                evalsList = np.vstack([evalsList, np.array(output)])
                            else:
                                for out in range(num_outputs):
                                    all_output_list[out][input][slice][k] = evalsList[im[0]][out]                       
                
                print(f"Number of total function evaluations: {num_evaluations + len(res)}")
                # Reshape all_output based on the actual number of elements per output
                all_output = np.array(all_output_list, dtype=object)

                all_output = np.reshape(
                    all_output, (num_outputs, d, -1, 2))  # Reshape to 4D

                # Calculate min and max for each output and input variable
                min_values = np.min(all_output, axis=3)
                max_values = np.max(all_output, axis=3)

                min_val = np.sort(min_values, axis=2)
                max_val = np.sort(max_values, axis=2)

                # Initialize bounds_input
                bounds_input = np.empty((num_outputs, d, n_disc, 2))

                for i in range(num_outputs):
                    for j in range(d):  # Iterate over input variables
                        # Merge min_values and max_values into bounds_input
                        for k in range(n_disc):
                            bounds_input[i, j, k, 0] = min_val[i, j, k]
                            bounds_input[i, j, k, 1] = max_val[i, j, k]

                        # Sort bounds_input along the last axis (k)
                        bounds_input[i, j, :, :] = np.sort(bounds_input[i, j, :, :], axis=-1)


                bounds = np.empty((num_outputs, 2, n_disc))  # Initialize bounds
                lower_bound = np.zeros((num_outputs, n_disc))  # Initialize lower_bound
                upper_bound = np.zeros((num_outputs, n_disc))  # Initialize upper_bound

                for i in range(num_outputs):
                    temp_bounds = []  # Temporary list for bounds

                    for k in range(n_disc):  # Iterate over input variables
                        # Impose the p-boxes for each input variable and store in temporary list
                        temp_bounds.append(imp(bounds_input[i, :, k, :]))

                    # Impose the p-boxes across all input variables for the current output
                    bounds[i, :, :] = np.array(temp_bounds).T

                    # Extract lower_bound and upper_bound from bounds
                    lower_bound[i, :] = bounds[i, 1, :]
                    upper_bound[i, :] = bounds[i, 0, :]

                if condensation is not None:
                    bounds = condense_bounds(bounds, condensation)

                results.raw_data['bounds'] = bounds
                results.raw_data['min'] = np.array([{'f': lower_bound[i, :]} for i in range(
                    num_outputs)])  # Initialize as a NumPy array
                results.raw_data['max'] = np.array([{'f': upper_bound[i, :]} for i in range(
                    num_outputs)])  # Initialize as a NumPy array

                if save_raw_data == 'yes':
                    print('No raw data provided for this method!')
                    # results.add_raw_data(f= all_output, x= x_combinations)

            elif save_raw_data == 'yes':  # If f is None and save_raw_data is 'yes'
                # results.add_raw_data(f= None, x= x_combinations)
                raise ValueError('No raw data provided for this method!')

            else:
                raise ValueError("No function is provided. Please provide a function!")
        
        case _: 
            raise ValueError(
                     "Invalid UP method! focused_discretisation_cauchy is under development.")

    return results


from __future__ import annotations
from typing import TYPE_CHECKING
from ...pba.intervals.number import Interval

import numpy as np
from typing import Callable
from pymoo.optimize import minimize
from pymoo.core.problem import Problem
from pymoo.algorithms.moo.nsga2 import NSGA2
from pymoo.algorithms.soo.nonconvex.ga import GA
from pymoo.core.callback import Callback
from pyuncertainnumber.propagation.utils import Propagation_results


def a(x):
    return np.asarray(x, dtype=float)


def genetic_optimisation_method(
    x_bounds: Interval | np.ndarray,
    f: Callable,
    results: Propagation_results = None,
    pop_size=1000,
    n_gen=100,
    tol=1e-3,
    n_gen_last=10,
    algorithm_type="NSGA2",
) -> Propagation_results:  # Specify return type
    """Performs both minimisation and maximisation using a genetic algorithm.

    args:
        x_bounds: Bounds for decision variables (NumPy array).
        f: Objective function to optimize.
        pop_size: Population size (int or array of shape (2,)).
        n_gen: Maximum number of generations (int or array of shape (2,)).
        tol: Tolerance for convergence check (float or array of shape (2,)).
        n_gen_last: Number of last generations to consider for convergence
                    (int or array of shape (2,)).
        algorithm_type: 'NSGA2' or 'GA' to select the optimisation algorithm
                        (str or array of shape (2,)).

    signature:
        genetic_optimisation_method(x_bounds: np.ndarray, f: Callable,
                                    pop_size=1000, n_gen=100, tol=1e-3,
                                    n_gen_last=10, algorithm_type="NSGA2") -> Propagation_results


    notes:
        It only handles a function which produces a single output.
        Refer to `pymoo.optimize` documentation for available options.

    returns:
        An `Propagation_results` object which contains:
            - 'un': UncertainNumber object(s) to represent the interval of the output.
            - 'raw_data' (dict): Dictionary containing raw data shared across output:
                    - 'x' (np.ndarray): Input values.
                    - 'f' (np.ndarray): Output values.
                    - 'min' (np.ndarray): Array of dictionaries for the function's output,
                              containing 'f' for the minimum of that output as well 'message', 'nit', 'nfev', 'final_simplex'.
                    - 'max' (np.ndarray): Array of dictionaries for the function's output,
                              containing 'f' for the maximum of that output as well 'message', 'nit', 'nfev', 'final_simplex'.
                    - 'bounds' (np.ndarray): 2D array of lower and upper bounds for the output.

    example:
        >>> # Example usage with different parameters for minimisation and maximisation
        >>> f = lambda x: x[0] + x[1] + x[2] # Example function
        >>> x_bounds = np.array([[1, 2], [3, 4], [5, 6]])
        >>> # Different population sizes for min and max
        >>> pop_size = np.array([500, 1500])
        >>> # Different number of generations
        >>> n_gen = np.array([50, 150])
        >>> # Different tolerances
        >>> tol = np.array([1e-2, 1e-4])
        >>> # Different algorithms
        >>> algorithm_type = np.array(["GA", "NSGA2"])
        >>> y = genetic_optimisation_method(x_bounds, f, pop_size=pop_size, n_gen=n_gen,
        >>>                                 tol=tol, n_gen_last=10, algorithm_type=algorithm_type)

    """

    if isinstance(x, Interval):
        x = x.to_numpy()

    # TODO move class definintion outside the function @ioannaioan
    class ProblemWrapper(Problem):
        def __init__(self, objective, **kwargs):
            super().__init__(n_obj=1, **kwargs)
            self.n_evals = 0
            self.objective = objective  # Store the objective ('min' or 'max')

        def _evaluate(self, x, out, *args, **kwargs):
            self.n_evals += len(x)

            # Evaluate the objective function for each individual separately
            # Apply f to each row of x
            res = np.array([f(x[i]) for i in range(len(x))])

            out["F"] = res if self.objective == "min" else -res

    class ConvergenceMonitor(Callback):
        def __init__(self, tol=1e-4, n_last=5):
            super().__init__()
            self.tol = tol  # Tolerance for
            self.n_last = n_last  # Number of last values to consider
            self.history = []  # Store the history of objective values
            self.n_generations = 0
            self.convergence_reached = False  # Flag to track convergence

        def notify(self, algorithm):
            self.n_generations += 1
            # Store best objective value
            self.history.append(algorithm.pop.get("F").min())

            # Check for convergence if enough values are available
            if len(self.history) >= self.n_last:
                last_values = self.history[-self.n_last :]
                # Calculate the range of the last 'n_last' values
                convergence_value = np.max(last_values) - np.min(last_values)
                if convergence_value <= self.tol and not self.convergence_reached:
                    self.convergence_message = (
                        "Convergence reached!"  # Store the message
                    )
                    print(self.convergence_message)
                    self.convergence_reached = True  # Set the flag to True
                    algorithm.termination.force_termination = True

    def run_optimisation(objective, pop_size, n_gen, tol, n_gen_last, algorithm_type):
        callback = ConvergenceMonitor(tol=tol, n_last=n_gen_last)
        problem = ProblemWrapper(
            objective=objective,
            n_var=x_bounds.shape[0],
            xl=x_bounds[:, 0],
            xu=x_bounds[:, 1],
        )
        algorithm = (
            GA(pop_size=pop_size)
            if algorithm_type == "GA"
            else NSGA2(pop_size=pop_size)
        )
        result = minimize(problem, algorithm, ("n_gen", n_gen), callback=callback)
        return (
            result,
            callback.n_generations,
            problem.n_evals,
            callback.convergence_message,
        )

    # Handle arguments that can be single values or arrays
    def handle_arg(arg):
        return np.array([arg, arg]) if not isinstance(arg, np.ndarray) else arg

    pop_size = handle_arg(pop_size)
    n_gen = handle_arg(n_gen)
    tol = handle_arg(tol)
    n_gen_last = handle_arg(n_gen_last)
    algorithm_type = handle_arg(algorithm_type)

    # --- Minimisation ---
    result_min, n_gen_min, n_iter_min, message_min = run_optimisation(
        "min", pop_size[0], n_gen[0], tol[0], n_gen_last[0], algorithm_type[0]
    )

    # --- Maximisation ---
    result_max, n_gen_max, n_iter_max, message_max = run_optimisation(
        "max", pop_size[1], n_gen[1], tol[1], n_gen_last[1], algorithm_type[1]
    )

    # Create a dictionary to store the results
    if results is None:
        results = Propagation_results()  # Create an instance of Propagation_results

    # Store ALL results in the results object with descriptions
    results.raw_data["min"] = np.append(
        results.raw_data["min"],
        {
            "x": result_min.X,
            "f": result_min.F,
            "message": message_min,
            "ngenerations": n_gen_min,
            "niterations": n_iter_min,
        },
    )

    results.raw_data["max"] = np.append(
        results.raw_data["min"],
        {
            "x": result_max.X,
            "f": -result_max.F,  # Negate the result for maximisation
            "message": message_max,
            "ngenerations": n_gen_max,
            "niterations": n_iter_max,
        },
    )

    results.raw_data["bounds"] = np.array([result_min.F[0], -result_max.F[0]])

    return results


# # Example usage with different parameters for minimisation and maximisation
# f = lambda x: x[0] + x[1] + x[2] # Example function
# x_bounds = np.array([[1, 2], [3, 4], [5, 6]])

# # Different population sizes for min and max
# pop_size = np.array([500, 1500])

# # Different number of generations
# n_gen = np.array([50, 150])

# # Different tolerances
# tol = np.array([1e-2, 1e-4])

# # Different algorithms
# algorithm_type = np.array(["GA", "NSGA2"])

# y = genetic_optimisation_method(x_bounds, f, pop_size=pop_size, n_gen=n_gen,
#                                         tol=tol, n_gen_last=10, algorithm_type=algorithm_type)

# # Print the results
# y.print()



import functools
from .uncertainNumber import UncertainNumber


def makeUN(func):
    """return from construct a Uncertain Number object"""

    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):
        construct = func(*args, **kwargs)
        return UncertainNumber.fromConstruct(construct)

    return wrapper_decorator



"""Constructors for Dempester-Shafer structures."""

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from .intervals.intervalOperators import make_vec_interval
from collections import namedtuple
import pyuncertainnumber.pba.aggregation as agg
from .intervals import Interval

dempstershafer_element = namedtuple("dempstershafer_element", ["interval", "mass"])
""" Named tuple for Dempster-Shafer elements.

note:
    - e.g. dempstershafer_element([0, 1], 0.5)
"""


class DempsterShafer:
    """Class for Dempester-Shafer structures.

    args:
    # TODO: add type hints
    # TODO: restructure the constructor needed
        - the `intervals` argument accepts wildcard vector intervals {list of list pairs or vec Interval};
        - masses (list): probability masses
    """

    def __init__(
        self, intervals: Interval | list[list] | np.ndarray, masses: list[float]
    ):

        self._intervals = make_vec_interval(intervals)
        self._masses = np.array(masses)

    def _create_DSstructure(self):
        return [
            dempstershafer_element(i, m) for i, m in zip(self._intervals, self._masses)
        ]

    def __repr__(self):
        return f"Dempster Shafer structure with {len(self._intervals)} focal elements"

    @property
    def structures(self):
        return self._create_DSstructure()

    @property
    def intervals(self):
        """Returns the Interval-typed focal elements of the Dempster-Shafer structure."""
        return self._intervals

    @property
    def focal_elements(self):
        """Returns the focal elements of the Dempster-Shafer structure."""
        return self._intervals

    @property
    def masses(self):
        return self._masses

    @property
    def naked_value(self):
        return np.round(np.sum(self._intervals.mid * self._masses), 3)

    def plot(self, style="box", ax=None, **kwargs):
        """for box type transform dss into a pbox and plot"""
        match style:
            case "box":
                dss_pbox = self.to_pbox()
                dss_pbox.plot(ax=ax, **kwargs)
            case "interval":
                plot_DS_structure(self.intervals, self.masses, ax=ax, **kwargs)

    def display(self, style="box", ax=None, **kwargs):
        self.plot(style=style, ax=ax, **kwargs)
        plt.show()

    def to_pbox(self):
        dss_pbox = agg.stacking(
            self.intervals,
            weights=self.masses,
            display=False,
            return_type="pbox",
        )
        return dss_pbox

    @classmethod
    def from_dsElements(cls, *ds_elements: dempstershafer_element):
        """Create a Dempster-Shafer structure from a list of Dempster-Shafer elements."""

        ds_elements = list(*ds_elements)
        intervals = [elem.interval for elem in ds_elements]
        masses = [elem.mass for elem in ds_elements]
        return cls(intervals, masses)


def plot_DS_structure(
    vec_interval: list[Interval],
    masses=None,
    offset=0.3,
    ax=None,
    **kwargs,
):
    """plot the intervals in a vectorised form

    args:
        vec_interval: vectorised interval objects
        masses: masses of the intervals
        offset: offset for display the masses next to the intervals
    """
    vec_interval = make_vec_interval(vec_interval)
    fig, ax = plt.subplots() if ax is None else (ax.figure, ax)
    for i, intl in enumerate(vec_interval):  # horizontally plot the interval
        ax.plot([intl.lo, intl.hi], [i, i], **kwargs)
        if masses is not None:
            ax.text(
                intl.hi + offset,
                i,
                f"{masses[i]:.2f}",
                verticalalignment="center",
                horizontalalignment="right",
            )
    ax.margins(x=0.2, y=0.1)
    ax.set_yticks([])
    return ax


def plot_DS_structure_with_labels(
    vec_interval: list[Interval],
    masses=None,
    offset=0.3,
    ax=None,
    **kwargs,
):
    """temp use: plot the intervals in a vectorised form

    args:
        vec_interval: vectorised interval objects
        masses: masses of the intervals
        offset: offset for display the masses next to the intervals
    """
    vec_interval = make_vec_interval(vec_interval)

    expert_l = ["a", "b", "c", "d"]

    fig, ax = plt.subplots() if ax is None else (ax.figure, ax)
    for i, intl in enumerate(vec_interval):  # horizontally plot the interval
        ax.plot([intl.lo, intl.hi], [i, i], label=f"expert {expert_l[i]}", **kwargs)

        if masses is not None:
            ax.text(
                intl.hi + offset,
                i,
                f"{masses[i]:.2f}",
                verticalalignment="center",
                horizontalalignment="right",
            )
    ax.margins(x=0.2, y=0.1)
    ax.set_yticks([])
    ax.legend()
    return ax



from numbers import Number
import numpy as np
import matplotlib.pyplot as plt
from statsmodels.distributions.copula.api import (
    FrankCopula,
    ClaytonCopula,
    GumbelCopula,
    GaussianCopula,
    StudentTCopula,
    IndependenceCopula,
)


class Dependency:
    """Dependecy class to specify copula models

    args:
        family (str): name of the copula family, e.g. "gaussian",
            "t", "frank", "gumbel", "clayton", "independence"
        params (Number): parameter of the copula, e.g. correlation for gaussian copula
            or degrees of freedom for t copula

    example:
        >>> from pyuncertainnumber import pba
        >>> pba.Dependency('gaussian', params=0.8)

    """

    # parameterisation init
    def __init__(self, family: str, params: Number):
        self.family = family
        self.params = params
        self._post_init_check()
        self._copula = self.copulas_dict.get(self.family)(params)

    copulas_dict = {
        "gaussian": GaussianCopula,
        "t": StudentTCopula,
        "frank": FrankCopula,
        "gumbel": GumbelCopula,
        "clayton": ClaytonCopula,
        "independence": IndependenceCopula,
    }

    def _post_init_check(self):
        supported_family_check(self.family)

    def __repr__(self):
        return f"copula: {self.family} with parameter {self.params}"

    def pdf(self, u):
        return self._copula.pdf(u)

    def cdf(self, u):
        return self._copula.cdf(u)

    def sample(self, n: int):
        """draws n samples in the U space"""
        return self._copula.rvs(n)

    def display(self, style="3d", ax=None):
        """show the PDF in the u space"""
        if style == "2d_pdf":
            self._copula.plot_pdf(ax=ax)
        elif style == "3d_cdf":
            grid_size = 100
            U, V = np.meshgrid(
                np.linspace(0, 1, grid_size), np.linspace(0, 1, grid_size)
            )
            Z = np.array(
                [
                    self._copula.cdf([U[i, j], V[i, j]])
                    for i in range(grid_size)
                    for j in range(grid_size)
                ]
            )
            Z = Z.reshape(grid_size, grid_size)
            pl_3d_copula(U, V, Z)
        else:
            raise ValueError("style must be '2d_pdf' or '3d_cdf'")

    def fit(self, data):
        return self._copula.fit_corr_param(data)


def supported_family_check(c):
    """check if copula family is supported"""
    if c not in {"gaussian", "t", "frank", "gumbel", "clayton", "independence"}:
        raise Exception("This copula model is not yet implemented")


def empirical_copula(data):
    """compute the empirical copula"""
    pass


def pl_3d_copula(U, V, Z):

    fig = plt.figure(figsize=(8, 8))
    ax = fig.add_subplot(111, projection="3d")
    ax.plot_surface(U, V, Z, cmap="viridis", edgecolor="none")
    ax.set_xlabel("u")
    ax.set_ylabel("v")
    ax.set_zlabel("C(u, v)")
    plt.show()



import numpy as np
import tqdm
from typing import Callable
from pyuncertainnumber.propagation.epistemic_uncertainty.cartesian_product import (
    cartesian,
)
from pyuncertainnumber.propagation.epistemic_uncertainty.extreme_point_func import (
    extreme_pointX,
)
from pyuncertainnumber.propagation.utils import Propagation_results


def extremepoints_method(
    x: np.ndarray, f: Callable, results: Propagation_results = None, save_raw_data=False
) -> Propagation_results:  # Specify return type
    """
        Performs uncertainty propagation using the Extreme Point Method for monotonic functions.
        This method estimates the bounds of a function's output by evaluating it at specific combinations of extreme values
        (lower or upper bounds) of the input variables. It is efficient for monotonic functions but might not be
        accurate for non-monotonic functions.

    args:
        x (np.ndarray): A 2D NumPy array where each row represents an input variable and
          the two columns define its lower and upper bounds (interval).
        f (Callable): A callable function that takes a 1D NumPy array of input values and
          returns the corresponding output(s).
        results (Propagation_results): The class to use for storing results (defaults to Propagation_results).
        save_raw_data (str, optional): Acts as a switch to enable or disable the storage of raw input data
            when a function (f)
            is not provided.
            - 'no': Returns an error that no function is provided.
            - 'yes': Returns the full arrays of unique input combinations.

    signature:
        extremepoints_method(x:np.ndarray, f:Callable, results:dict, save_raw_data = 'no') -> dict

    raises:
        ValueError if no function is provided and save_raw_data is 'no'.

    return:
        Returns `Propagation_results` object(s) containing:
            - 'un': UncertainNumber object(s) to characterise the interval(s) of the output(s).
            - 'raw_data' (dict): Dictionary containing raw data shared across output(s):
                    - 'x' (np.ndarray): Input values.
                    - 'f' (np.ndarray): Output values.
                    - 'min' (np.ndarray): Array of dictionaries, one for each output,
                              containing 'f' for the minimum of that output.
                    - 'max' (np.ndarray): Array of dictionaries, one for each output,
                              containing 'f' for the maximum of that output.
                    - 'bounds' (np.ndarray): 2D array of lower and upper bounds for each output.

    Example:
        # Example usage with different parameters for minimization and maximization
        >>> f = lambda x: x[0] + x[1] + x[2]  # Example function
        >>> # Determine input parameters for function and method
        >>> x_bounds = np.array([[1, 2], [3, 4], [5, 6]])
        >>> # Call the method
        >>> y = extremepoint_method(x_bounds, f)
        >>> # print results
        >>> y.print()

    """
    if results is None:
        results = Propagation_results()  # Create an instance of Propagation_results

    # create an array with the unique combinations of all intervals
    X = cartesian(*x)

    d = X.shape[1]  # Number of dimensions
    inds = np.array([1] + [2**i + 1 for i in range(d)])  # Generate indices
    # Select rows based on indices (adjusting for 0-based indexing)
    Xeval = X[inds - 1]

    # propagates the epistemic uncertainty through extremepoints
    if f is not None:

        # Simulate function for the selected subset
        all_output = []
        for c in tqdm.tqdm(Xeval, desc="Number of function evaluations"):
            output = f(c)
            all_output.append(output)

        # Determine the number of outputs from the first evaluation
        try:
            num_outputs = len(all_output[0])
        except TypeError:
            num_outputs = 1  # If f returns a single value

        # Convert all_output to a NumPy array with the correct shape
        all_output = np.array(all_output).reshape(-1, num_outputs)

        # Calculate signs
        part_deriv_sign = np.zeros((num_outputs, d))
        Xsign = np.zeros((2 * num_outputs, d))
        for i in range(num_outputs):
            # Calculate signs based on initial output values
            part_deriv_sign[i] = np.sign(all_output[1:, i] - all_output[0, i])[::-1]

            # Calculate extreme points
            Xsign[2 * i : 2 * (i + 1), :] = extreme_pointX(x, part_deriv_sign[i])

        # Store and reuse evaluations
        evaluated_points = {}
        lower_bound = np.full(num_outputs, np.inf)  # Initialize with inf
        upper_bound = np.full(num_outputs, -np.inf)  # Initialize with -inf
        min_indices = np.zeros((d, num_outputs))
        max_indices = np.zeros((d, num_outputs))
        num_evaluations = 0  # Initialize evaluation counter

        for i in range(num_outputs):
            for j in range(2):  # For lower and upper bounds
                key = tuple(Xsign[2 * i + j, :])
                if key not in evaluated_points:
                    evaluated_points[key] = f(Xsign[2 * i + j, :])
                    num_evaluations += 1
                if num_outputs == 1:
                    output = evaluated_points[key]

                else:
                    output = evaluated_points[key][i]

                # Determine if this is the lower or upper bound for this output
                if output < lower_bound[i]:
                    lower_bound[i] = output
                    min_indices[:, i] = Xsign[2 * i + j, :]
                if output > upper_bound[i]:
                    upper_bound[i] = output
                    max_indices[:, i] = Xsign[2 * i + j, :]

        print(f"Number of total function evaluations: {num_evaluations + len(Xeval)}")
        # Convert to 2D arrays (if necessary) and append
        for i in range(num_outputs):
            results.raw_data["min"] = np.append(
                results.raw_data["min"], {"x": min_indices[:, i], "f": lower_bound[i]}
            )
            results.raw_data["max"] = np.append(
                results.raw_data["max"], {"x": max_indices[:, i], "f": upper_bound[i]}
            )
            results.raw_data["bounds"] = (
                np.vstack(
                    [
                        results.raw_data["bounds"],
                        np.array([lower_bound[i], upper_bound[i]]),
                    ]
                )
                if results.raw_data["bounds"].size
                else np.array([lower_bound[i], upper_bound[i]])
            )

        results.add_raw_data(part_deriv_sign=part_deriv_sign)
        results.raw_data["x"] = Xeval
        results.raw_data["f"] = all_output

    elif save_raw_data:  # If f is None and save_raw_data is 'yes'
        # Store Xeval in raw_data['x'] even if f is None
        results.add_raw_data(x=Xeval)

    else:
        raise ValueError(
            "No function is provided. Select save_raw_data = 'yes' to save the input combinations"
        )

    return results  # only the interval



from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np
from .intervals.intervalOperators import make_vec_interval
from .utils import reweighting
from .ecdf import eCDF_bundle, get_ecdf
from .intervals import Interval
import functools
from numbers import Number
from ..decorator import UNtoUN
from .pbox_abc import Staircase, convert_pbox
from .utils import expose_functions_as_public


if TYPE_CHECKING:
    from .pbox_abc import Pbox
    from .dss import DempsterShafer


# envelope
def _envelope(*l_constructs: Pbox | DempsterShafer | Number) -> Staircase:
    """calculates the envelope of uncertain number constructs

    args:
        l_constructs (list): the components, constructs only, on which the envelope operation applied on.

    returns:
        the envelope of the given arguments,  either a p-box or an interval.
    """

    def binary_env(p1, p2):
        return p1.env(p2)

    xs = [convert_pbox(x) for x in l_constructs]
    return functools.reduce(binary_env, xs)


# imposition
def _imposition(*l_constructs: Staircase | DempsterShafer | Number) -> Staircase:
    """Returns the imposition/intersection of the list of p-boxes

    args:
        - l_constructs (list): a list of UN objects to be mixed

    returns:
        - Pbox

    note:
        - #TODO verfication needed for the base function `p1.imp(p2)`
    """

    def binary_imp(p1, p2):
        return p1.imp(p2)

    xs = [convert_pbox(x) for x in l_constructs]
    return functools.reduce(binary_imp, xs)


# mixture
def _stochastic_mixture(*l_constructs, weights=None, display=False, **kwargs):
    """it could work for either Pbox, distribution, DS structure or Intervals

    args:
        - l_constructs (list): list of constructs of uncertain number
        - weights (list): list of weights
        - display (Boolean): boolean for plotting
    # TODO mix types later
    note:
        - currently only accepts same type objects
    """

    from .pbox_abc import Pbox
    from .dss import DempsterShafer
    from .intervals import Interval

    if isinstance(l_constructs[0], Interval | list):
        return stacking(l_constructs, weights=weights, display=display, **kwargs)
    elif isinstance(l_constructs[0], Pbox):
        return mixture_pbox(*l_constructs, weights, display=display)
    elif isinstance(l_constructs[0], DempsterShafer):
        return mixture_ds(*l_constructs, display=display)


# * --------------- construct levels ----------------- * #


def stacking(
    vec_interval: Interval | list[Interval],
    *,
    weights=None,
    display=False,
    ax=None,
    return_type="pbox",
    **kwargs,
) -> Pbox:
    """stochastic mixture operation of Intervals with probability masses

    args:
        - l_constructs (list): list of constructs of uncertain numbers
        - weights (list): list of weights
        - display (Boolean): boolean for plotting
        - return_type (str): {'pbox' or 'ds' or 'bounds'}

    return:
        - the left and right bound F in `eCDF_bundlebounds` by default
        but can choose to return a p-box

    note:
        - it takes a list of intervals or a single vectorised interval, which is
        a different signature compared to the other aggregation functions.
        - together the interval and masses, it can be deemed that all the inputs
        required is jointly a DS structure
    """
    from .pbox_abc import Staircase
    from .dss import DempsterShafer
    from .ecdf import plot_two_eCDF_bundle

    vec_interval = make_vec_interval(vec_interval)
    q1, p1 = get_ecdf(vec_interval.lo, weights)
    q2, p2 = get_ecdf(vec_interval.hi, weights)

    cdf1 = eCDF_bundle(q1, p1)
    cdf2 = eCDF_bundle(q2, p2)

    if display:
        plot_two_eCDF_bundle(cdf1, cdf2, ax=ax, **kwargs)

    match return_type:
        case "pbox":
            return Staircase.from_CDFbundle(cdf1, cdf2)
        case "dss":
            return DempsterShafer(intervals=vec_interval, masses=weights)
        case "cdf":
            return cdf1, cdf2
        case _:
            raise ValueError("return_type must be one of {'pbox', 'dss', 'cdf'}")


def mixture_pbox(*l_pboxes, weights=None, display=False) -> Pbox:

    if weights is None:
        N = len(l_pboxes)
        weights = np.repeat(1 / N, N)  # equal weights
    else:
        weights = np.array(weights) if not isinstance(weights, np.ndarray) else weights
        weights = weights / sum(weights)  # re-weighting

    lcdf = np.sum([p.left * w for p, w in zip(l_pboxes, weights)], axis=0)
    ucdf = np.sum([p.right * w for p, w in zip(l_pboxes, weights)], axis=0)
    pb = Pbox(left=lcdf, right=ucdf)
    if display:
        pb.display(style="band")
    return pb


def mixture_ds(*l_ds, display=False) -> DempsterShafer:
    """mixture operation for DS structure"""

    from .dss import DempsterShafer

    intervals = np.concatenate([ds.intervals.to_numpy() for ds in l_ds], axis=0)
    # TODO check the duplicate intervals
    # assert sorted(intervals) == np.unique(intervals), "intervals replicate"
    masses = reweighting([ds.masses for ds in l_ds])
    return DempsterShafer(intervals, masses)


def env_ecdf(data, ret_type="pbox", ecdf_choice="canonical"):
    """nonparametric envelope function

    arrgs:
        data (array): Each row represents a distribution, on which the envelope operation applied.
        ret_type (str): {'pbox' or 'cdf'}
            - default is pbox
            - cdf is the CDF bundle
        ecdf_choice (str): {'canonical' or 'staircase'}

    note:
        envelope on a set of empirical CDFs
    """
    from .ecdf import ecdf, get_ecdf

    ecdf_func = get_ecdf if ecdf_choice == "canonical" else ecdf

    # assume each row as a sample and eCDF
    q_list = []
    for l in range(data.shape[0]):
        dd, pp = ecdf_func(np.squeeze(data[l]))
        q_list.append(dd)

    # return the q lower bound which is the upper probability bound
    q_arr = np.array(q_list)
    l_bound = np.min(q_arr, axis=0)
    u_bound = np.max(q_arr, axis=0)

    if ret_type == "pbox":
        return Staircase(left=l_bound, right=u_bound)
    elif ret_type == "cdf":
        return eCDF_bundle(l_bound, pp), eCDF_bundle(u_bound, pp)


def env_ecdf_sep(*ecdfs, ret_type="pbox", ecdf_choice="canonical"):
    """nonparametric envelope function for separate empirical CDFs"""

    data = np.array(ecdfs)
    return env_ecdf(data, ret_type=ret_type, ecdf_choice=ecdf_choice)


# exposed APIs


# envelope = UNtoUN(_envelope)

# Mapping: public API name -> private function
api_map = {
    "envelope": _envelope,
    "imposition": _imposition,
    "stochastic_mixture": _stochastic_mixture,
}

expose_functions_as_public(api_map, UNtoUN)

__all__ = list(api_map.keys())
__all__.extend(
    [
        "stacking",
        "mixture_pbox",
        "mixture_ds",
        "env_ecdf",
        "env_ecdf_sep",
    ]
)



from __future__ import annotations
from typing import TYPE_CHECKING

"""leslie's general bound to bound implementation"""

from ...pba.intervals.number import Interval
from ...pba.intervals.intervalOperators import make_vec_interval
import numpy as np

if TYPE_CHECKING:
    from ...pba.intervals import Interval


# TODO: integrate GA and BO implementations
# TODO: add discussion of `func` signature (args, collection, matrix) in the notes section
def b2b(
    vecs: Interval | list[Interval],
    func,
    interval_strategy=None,
    style=None,
    n_sub=None,
    **kwargs,
) -> Interval:
    """General implementation of interval propagation through a function:

    .. math::
        Y = g(I_{x1}, I_{x2}, ..., I_{xn})

    where :math:`I_{x1}, I_{x2}, ..., I_{xn}` are intervals.

    In a general case, the function :math:`g` is not necessarily monotonic and :math:`g` may be a black-box model.
    Optimisation to the rescue and two of them particularly: GA and BO.

    args:
        vecs (Interval): a vector Interval or a list or tuple of scalar Interval
        func (callable): performance or response function or a black-box model as in subprocess.
            Expect 2D inputs therefore `func` shall have the matrix signature. See Notes for additional details.
        interval_strategy (str): the interval_strategy used for interval propagation

            - 'endpoints': only the endpoints

            - 'ga': genetic algorithm

            - 'bo': bayesian optimisation

            - 'diret': direct apply function (the default)

        style (str): the style only used for subinterval propagation
        **kwargs: additional keyword arguments to be passed to the function


    tip:
        This serves as a top-level func as `epistemic_propagation()`.

    caution:
        ``interval_strategy`` suggests the method of interval propagation whhile ``style`` is the style of subinterval propagation.

    returns:
        Interval: the low and upper bound of the response
    """

    vec_itvl = make_vec_interval(vecs)
    match interval_strategy:
        case "endpoints":
            return endpoints(vec_itvl, func)
        case "subinterval":
            return subinterval_method(
                vec_itvl, func, style=style, n_sub=n_sub, **kwargs
            )
        case "ga":
            pass
        case "bo":
            pass
        case "direct":
            return func(vec_itvl)
        case _:
            raise NotImplementedError(
                f"Method {interval_strategy} is not supported yet."
            )


def endpoints(vec_itvl: Interval, func) -> Interval:
    """Implementation of endpoints method

    args:
        vec_itvl (Interval): a vector type Interval object
        func (callable): the function to be evaluated. See notes about function signature.

    note:
        The function `func` is expected to accept a 2D numpy array of shape :math:`(2^d, d)` where `d` is the dimension of the vector interval.
        Therefore, the function  should have a vectorised signature, as opposted to taking individual scalar inputs.


    Example:
        >>> from pyuncertainnumber import pba
        >>> v = pba.I([1, 2], [3, 4])  # a vector interval with two dimensions
        >>> def bar(x): return x[:, 0] ** 3 + x[:, 1] + 5
        >>> endpoints(v, bar)
        Interval(8.0, 36.0)
    """

    v_np = vec_itvl.to_numpy()
    rows = np.vsplit(v_np, v_np.shape[0])
    arr = vec_cartesian_product(*rows)
    # print(arr.shape)  # array of shape (2**n, 2)

    response = func(arr)  # func on each row of combination of endpoints
    min_response = np.min(response)
    max_response = np.max(response)
    return Interval(min_response, max_response)


def subinterval_method(
    vec_itvl: Interval, func, style=None, n_sub=None, parallel=False
) -> Interval:
    # TODO parallel subinterval
    """Implmentation of subinterval method which splits subintervals and reconstitutes later.

    Args:
        vec_itvl (Interval): a vector type Interval object
        func (callable): the function to be evaluated. See notes about function signature.
        n_sub (int): number of subintervals
        style (str): the style used for interval propagation which shall be compatible with the response function signature.

            - 'direct': direct apply function

            - 'endpoints': only the endpoints are propagated

    note:
        There are some subtleties about function signature. For ``endpoints`` style, the function `func` should have a vectorised signature
        as it is expecting a 2D numpy array, whereas for the `direct` style, it is expecting to take individual scalar inputs

    Example:
        >>> from pyuncertainnumber import pba
        >>> v = pba.I([1, 2], [3, 4])
        >>> def bar(x): return x[0] ** 3 + x[1] + 5
        >>> b2b.subinterval_method(vec_itvl=v, func=bar, style='direct', n_sub=2)
        >>> def bar_vec(x): return x[:, 0] ** 3 + x[:, 1] + 5
        >>> b2b.subinterval_method(vec_itvl=v, func=bar_vec, style='endpoints', n_sub=2)
    """
    from pyuncertainnumber.pba.intervals.methods import subintervalise, reconstitute

    if style is None:
        raise ValueError("style must be chosen within {'direct', 'endpoints'}.")
    if n_sub is None:
        raise ValueError("Number of subintervals n_sub must be provided.")

    sub = subintervalise(vec_itvl, n_sub)
    if style == "direct":
        row_n = sub.shape[0]
        return reconstitute([func(sub[IND]) for IND in range(row_n)])
    elif style == "endpoints":
        return reconstitute([endpoints(sub[i], func) for i in range(len(sub))])


def vec_cartesian_product(*arrays):
    """a vectorised version of the cartesian product

    args:
        arrays: a couple of 1D np.ndarray objects
    """
    grids = np.meshgrid(*arrays, indexing="ij")
    # TODO nice imple but not directrly working with vec Interval object yet
    stacked = np.stack(grids, axis=-1)
    return stacked.reshape(-1, len(arrays))


def i_cartesian_product(a, b):
    """a vectorisation of the interval cartesian product

    todo:
        extend to multiple input arguments
    """
    from pyuncertainnumber import pba

    # Extract bounds
    a_lower = a.lo[:, np.newaxis]  # (2, 1)
    a_upper = a.hi[:, np.newaxis]  # (2, 1)
    b_lower = b.lo[np.newaxis, :]  # (1, 2)
    b_upper = b.hi[np.newaxis, :]  # (1, 2)

    # Broadcast to shape (2, 2)
    cart_lower = np.stack(
        [
            a_lower.repeat(b_lower.shape[1], axis=1),
            np.tile(b_lower, (a_lower.shape[0], 1)),
        ],
        axis=-1,
    )  # shape (2, 2, 2)

    cart_upper = np.stack(
        [
            a_upper.repeat(b_upper.shape[1], axis=1),
            np.tile(b_upper, (a_upper.shape[0], 1)),
        ],
        axis=-1,
    )  # shape (2, 2, 2)

    # Reshape to flat list of interval pairs
    flat_lower = cart_lower.reshape(-1, 2)  # shape (4, 2)
    flat_upper = cart_upper.reshape(-1, 2)  # shape (4, 2)

    return pba.I(lo=flat_lower, hi=flat_upper)


"""
POOL:
            except IndexError as e:
                if "too many indices for array" in str(e):
                    print("2D inputs expected but 1D presented:", e)
                    return endpoints(vec_itvl[None, :], func)
                else:
                    raise  # Re-raise if it's a different IndexError
"""



from __future__ import annotations
from typing import TYPE_CHECKING
from ...pba.intervals.number import Interval

import numpy as np
from rich.progress import track
from typing import Callable
from .cartesian_product import cartesian
from ..utils import Propagation_results


def subinterval_method(
    x: np.ndarray,
    f: Callable,
    save_raw_data=False,
    n_sub: np.array = 10,
    results: Propagation_results = None,
) -> Propagation_results:  # Specify return type
    """The subinterval reconstitution method.

    args:
        x (nd.array): A 2D NumPy array where each row represents an input variable and the two columns
            define its lower and upper bounds (interval).
        f (callable): A callable function that takes a 1D NumPy array of input values and returns the
            corresponding output(s).
        results (Propagation_results): The class to use for storing results (defaults to Propagation_results).
        n_sub (nd.array): A scalar (integer) or a 1D NumPy array specifying the number of subintervals for
            each input variable.
            - If a scalar, all input variables are divided into the same number of subintervals (defaults 3 divisions).
            - If an array, each element specifies the number of subintervals for the
               corresponding input variable.
        save_raw_data (str, optional): Acts as a switch to enable or disable the storage of raw input data when a function (f)
            is not provided.
            - 'no': Returns an error that no function is provided.
            - 'yes': Returns the full arrays of unique input combinations.

    signature:
        subinterval_method(x:np.ndarray, f:Callable, n_sub:np.array ...) -> Propagation_results

    notes:
        - The function assumes that the intervals in `x` represent epistemic uncertainties in the input.
        - The subinterval reconstitution method subdivides the input intervals into smaller subintervals
          to accommodate for the presence of non-monotonic trends in the function output(s).
        - The subintervals for the input can vary in number.
        - The computational cost increases exponentially with the number of input variables
          and the number of subintervals per variable.
        - The `f` function can return multiple outputs.

    raises:
        ValueError if no function is provided and save_raw_data is 'no'.

    returns:
        `Propagation_results` object(s) containing:
            - 'un': UncertainNumber object(s) to characterise the interval(s) of the output(s).
            - 'raw_data' (dict): Dictionary containing raw data shared across output(s):
                    - 'x' (np.ndarray): Input values.
                    - 'f' (np.ndarray): Output values.
                    - 'min' (np.ndarray): Array of dictionaries, one for each output,
                              containing 'f' for the minimum of that output.
                    - 'max' (np.ndarray): Array of dictionaries, one for each output,
                              containing 'f' for the maximum of that output.
                    - 'bounds' (np.ndarray): 2D array of lower and upper bounds for each output.

    example:
        >>> #Define input intervals
        >>> x = np.array([[1, 2], [3, 4], [5, 6]])
        >>> # Define the function
        >>> f = lambda x: x[0] + x[1] + x[2]
        >>> # Run sampling method with n = 2
        >>> y = subinterval_method(x, f, n_sub, save_raw_data = 'yes')
        >>> # Print the results
        >>> y.print()
    """
    if isinstance(x, Interval):
        x = x.to_numpy()

    if results is None:
        results = Propagation_results()  # Create an instance of Propagation_results

    # Create a sequence of values for each interval based on the number of divisions provided
    # The divisions may be the same for all intervals or they can vary.
    m = x.shape[0]

    if type(n_sub) == int:  # All inputs have identical division
        total = (n_sub + 1) ** m
        Xint = np.zeros((0, n_sub + 1), dtype=object)
        for xi in x:
            new_X = np.linspace(xi[0], xi[1], n_sub + 1)
            Xint = np.concatenate((Xint, new_X.reshape(1, n_sub + 1)), axis=0)
    else:  # Different divisions per input
        total = 1
        Xint = []
        for xc, c in zip(x, range(len(n_sub))):
            total *= n_sub[c] + 1
            new_X = np.linspace(xc[0], xc[1], n_sub[c] + 1)

            Xint.append(new_X)

        Xint = np.array(Xint, dtype=object)
    # create an array with the unique combinations of all subintervals
    X = cartesian(*Xint)

    # propagates the epistemic uncertainty through subinterval reconstitution
    if f is not None:
        all_output = np.array(
            [f(xi) for xi in track(X, description="Function evaluations")]
        )

        try:
            num_outputs = len(all_output[0])
        except TypeError:
            num_outputs = 1  # If f returns a single value

        # Reshape all_output to a 2D array (Corrected)
        all_output = np.array(all_output).reshape(-1, num_outputs)

        # Calculate lower and upper bounds for each output
        lower_bound = np.min(all_output, axis=0)
        upper_bound = np.max(all_output, axis=0)

        # Find indices of minimum and maximum values for each output
        min_indices = np.array(
            [
                np.where(all_output[:, i] == lower_bound[i])[0]
                for i in range(num_outputs)
            ]
        )
        max_indices = np.array(
            [
                np.where(all_output[:, i] == upper_bound[i])[0]
                for i in range(num_outputs)
            ]
        )

        # Convert to 2D arrays (if necessary) and append
        for i in range(num_outputs):
            results.raw_data["min"] = np.append(
                results.raw_data["min"], {"x": X[min_indices[i]], "f": lower_bound[i]}
            )
            results.raw_data["max"] = np.append(
                results.raw_data["max"], {"x": X[max_indices[i]], "f": upper_bound[i]}
            )
            results.raw_data["bounds"] = (
                np.vstack(
                    [
                        results.raw_data["bounds"],
                        np.array([lower_bound[i], upper_bound[i]]),
                    ]
                )
                if results.raw_data["bounds"].size
                else np.array([lower_bound[i], upper_bound[i]])
            )

        results.raw_data["f"] = all_output
        results.raw_data["x"] = X

    elif save_raw_data:  # If f is None and save_raw_data is 'yes'
        results.add_raw_data(x=X)
    else:
        print(
            "No function is provided. Select save_raw_data = 'yes' to save the input combinations"
        )

    return results


# # Example usage with different parameters for minimization and maximization
# f = lambda x: x[0] + x[1] + x[2]  # Example function

# # Determine input parameters for function and method
# x_bounds = np.array([[1, 2], [3, 4], [5, 6]])
# n=2
# # Call the method
# y = subinterval_method(x_bounds, f=None, n_sub=n, save_raw_data = 'yes')



from bayes_opt import BayesianOptimization
from bayes_opt import acquisition
import numpy as np
import inspect
from functools import partial
import warnings


class BayesOpt:
    """Bayesian Optimization class for the uncertain number"""

    # TODO add a descriptor for `task`
    def __init__(
        self,
        f,
        xc_bounds,
        dimension,
        task="maximisation",
        num_explorations=100,
        num_iterations=100,
        acquisition_function=None,
    ):

        self.task = task  # either minimisation or maximisation
        self.num_explorations = num_explorations  # initial exploration points
        self.num_iterations = num_iterations
        self.xc_bounds = xc_bounds
        self.dimension = dimension
        self.acquisition_function = (
            acquisition_function
            if acquisition_function
            else acquisition.UpperConfidenceBound(kappa=10.0)
        )
        self.f = f  # the function to be optimised

    @property
    def f(self):
        """return the function to be optimised"""
        return self._f

    @f.setter
    def f(self, f):
        # step 1: dimension check
        if not check_argument_count(f) == "Single argument":
            warnings.warn(
                "The function to be optimised should have a single argument",
                category=RuntimeWarning,
            )
        if self.dimension > 1 and (check_argument_count(f) == "Single argument"):
            f = partial(transform_func, fb=f)
        # step 2: flip check by the task
        # the first flip is to make the function minimisation
        if self.task == "maximisation":
            self._f = f
        elif self.task == "minimisation":
            self._f = lambda *args, **kwargs: -f(*args, **kwargs)

    def get_results(self):
        """inspect the results, to save or not"""

        self._optimal_dict = {}
        # TODO serialise the dict, plus the undering GP model

        if self.task == "maximisation":
            bo_all_dict = {
                "Xc_params": self.optimizer.space.params.tolist(),
                "target_array": self.optimizer.space.target.tolist(),
                "optimal_Xc": list(self.optimizer.max["params"].values()),
                "optimal_target": self.optimizer.max["target"],
            }
        elif self.task == "minimisation":  # the second flip
            target_arr = self.optimizer.space.target.copy()
            target_arr[:] *= -1

            optimal_index = np.argmin(target_arr)
            optimal_target = np.min(target_arr)
            optimal_Xc = self.optimizer.space.params[optimal_index]

            bo_all_dict = {
                "Xc_params": self.optimizer.space.params.tolist(),
                "target_array": target_arr.tolist(),
                "optimal_Xc": optimal_Xc.tolist(),
                "optimal_target": optimal_target.tolist(),
            }

        self._optimal_dict["xc"] = bo_all_dict["optimal_Xc"]
        self._optimal_dict["target"] = bo_all_dict["optimal_target"]
        self._all_results = bo_all_dict

    def run(self, **kwargs):

        self.optimizer = BayesianOptimization(
            f=self.f,
            pbounds=self.xc_bounds,
            acquisition_function=self.acquisition_function,
            random_state=42,
            allow_duplicate_points=True,
            **kwargs,
        )

        try:
            # initial exploration of the design space
            self.optimizer.maximize(
                init_points=self.num_explorations,
                n_iter=0,
            )
        except:
            pass

        # * _________________ run the BO iterations to get the optimal Xc
        for _ in range(self.num_iterations):
            next_point = self.optimizer.suggest()
            target = self._f(**next_point)
            self.optimizer.register(params=next_point, target=target)
            # print(target, next_point)

        # * _________________ compile the results
        self.get_results()

    @property
    def optimal(self):
        return self._optimal_dict

    @property
    def optimal_xc(self):
        """return the optimal xc"""
        return self._optimal_dict["xc"]

    @property
    def optimal_target(self):
        """return the optimal target"""
        return self._optimal_dict["target"]


def check_argument_count(func):
    # Get the function signature
    sig = inspect.signature(func)
    # Count the number of non-default parameters
    param_count = len(
        [p for p in sig.parameters.values() if p.default == inspect.Parameter.empty]
    )
    return "Single argument" if param_count == 1 else "Multiple arguments"


def transform_func(fb, **kwargs):
    args = [p for p in kwargs.values()]
    return fb(args)



import numpy as np
from geneticalgorithm import geneticalgorithm as ga


class GA:
    """high-level interface for the genetic algorithm"""

    # TODO add a descriptor for `task`
    def __init__(self, f, task, dimension, varbound):
        """
        args:
            f (callable): function to be optimized
        """
        self.f = f
        self.task = task
        self.dimension = dimension
        self.varbound = varbound
        self.setup()

    def setup(self):
        if self.task == "maximisation":
            self.flip_f = lambda *args, **kwargs: -self.f(*args, **kwargs)
            self._f = self.flip_f
        elif self.task == "minimisation":
            self._f = self.f
        else:
            raise ValueError("task should be either 'maximisation' or 'minimisation'")

    def get_results(self):
        """display the results of the optimization"""
        # direct result from GA
        self.output = self.model.output_dict.copy()

        if self.task == "maximisation":
            self.output["function"] = np.absolute(self.output["function"])
        elif self.task == "minimisation":
            pass

        self._optimal_dict = {}
        self._optimal_dict["xc"] = self.output["variable"]
        self._optimal_dict["target"] = self.output["function"]
        self._all_results = self.output
        # hint: self.model.output_dict["variable"], self.model.output_dict["function"]

    def run(self, algorithm_param=None, **kwargs):
        """run the genetic algorithm"""
        if algorithm_param is not None:
            self.model = ga(
                function=self._f,
                dimension=self.dimension,
                variable_type="real",
                variable_boundaries=self.varbound,
                algorithm_parameters=self.algorithm_param,
                function_timeout=int(1e6),
                **kwargs,
            )
        else:
            self.model = ga(
                function=self._f,
                dimension=self.dimension,
                variable_type="real",
                variable_boundaries=self.varbound,
                function_timeout=int(1e6),
                **kwargs,
            )
        self.model.run()
        self.get_results()

    @property
    def optimal(self):
        return self._optimal_dict

    @property
    def optimal_xc(self):
        """return the design variable that gives the optimal function value"""
        try:
            return self.model.output_dict["variable"].item()
        except AttributeError:
            raise AttributeError("You need to run the model first.")

    @property
    def optimal_target(self):
        """return the optimal target value"""
        return self._optimal_dict["target"]



import numpy as np
from typing import Callable
from scipy.optimize import minimize
from pyuncertainnumber.propagation.utils import Propagation_results


def local_optimisation_method(x: np.ndarray, f: Callable, x0: np.ndarray = None,
                              results: Propagation_results = None,
                              tol_loc: np.ndarray = None, options_loc: dict = None,
                              *, method_loc='Nelder-Mead') -> Propagation_results:  # Specify return type
    """  
        Performs local optimization to find both the minimum and maximum values of a given function, within specified bounds.

    args:
        x (np.ndarray): A 2D NumPy array where each row represents an input variable and 
                        the two columns define its lower and upper bounds (interval).
        f (Callable): The objective function to be optimized. It should take a 1D NumPy array 
                      as input and return a scalar value.
        results (Propagation_results, optional): An object to store propagation results.
                                            Defaults to None, in which case a new
                                            `Propagation_results` object is created.
        x0 (np.ndarray, optional): A 1D or 2D NumPy array representing the initial guess for the 
                                    optimization. 
                                    - If x0 has shape (n,), the same initial values are used for both 
                                      minimization and maximization.
                                    - If x0 has shape (2, n), x0[0, :] is used for minimization and 
                                      x0[1, :] for maximization.
                                    If not provided, the midpoint of each variable's interval is used.
                                    Defaults to None.
        tol_loc (np.ndarray, optional): Tolerance for termination.
                                        - If tol_loc is a scalar, the same tolerance is used for both 
                                          minimization and maximization.
                                        - If tol_loc is an array of shape (2,), tol_loc[0] is used for 
                                          minimization and tol_loc[1] for maximization.
                                        Defaults to None.
        options_loc (dict, optional): A dictionary of solver options. 
                                      - If options_loc is a dictionary, the same options are used for 
                                        both minimization and maximization.
                                      - If options_loc is a list of two dictionaries, options_loc[0] 
                                        is used for minimization and options_loc[1] for maximization.
                                      Refer to `scipy.optimize.minimize` documentation for available 
                                      options. Defaults to None.
        method_loc (str, optional): The optimization method to use (e.g., 'Nelder-Mead', 'COBYLA'). 
                                    Defaults to 'Nelder-Mead'.

   signature:
        local_optimisation_method(x:np.ndarray, f:Callable,  results = None,
                              *, x0:np.ndarray = None,  
                              tol_loc:np.ndarray = None, options_loc: dict = None, method_loc = 'Nelder-Mead') -> dict

    notes:
        This function utilizes the `scipy.optimize.minimize` function to perform local optimization. 
        Refer to `scipy.optimize.minimize` documentation for available options.
        It only handles a function which produces a single output.
    
    returns:
        An `Propagation_results` object which contains:
            - 'un': UncertainNumber object(s) to represent the interval of the output.
            - 'raw_data' (dict): Dictionary containing raw data shared across output:
                    - 'x' (np.ndarray): Input values.
                    - 'f' (np.ndarray): Output values.
                    - 'min' (np.ndarray): Array of dictionaries for the function's output,
                              containing 'f' for the minimum of that output as well 'message', 'nit', 'nfev', 'final_simplex'.
                    - 'max' (np.ndarray): Array of dictionaries for the function's output,
                              containing 'f' for the maximum of that output as well 'message', 'nit', 'nfev', 'final_simplex'.
                    - 'bounds' (np.ndarray): 2D array of lower and upper bounds for the output.
    
    example:
        >>> f = lambda x: x[0] + x[1] + x[2]  # Example function
        >>> x_bounds = np.array([[1, 2], [3, 4], [5, 6]])
        >>> # Initial guess (same for min and max)
        >>> x0 = np.array([1.5, 3.5, 5.5])  
        >>> # Different tolerances for min and max
        >>> tol_loc = np.array([1e-4, 1e-6])  
        >>> # Different options for min and max
        >>> options_loc = [
        >>>     {'maxiter': 100},  # Options for minimization
        >>>     {'maxiter': 1000}  # Options for maximization
        >>>     ]
        >>> # Perform optimization
        >>> y = local_optimisation_method(x_bounds, f, x0=x0, tol_loc=tol_loc, 
        >>>                                     options_loc=options_loc)

    """

    bounds = [(var[0], var[1]) for var in x]

    def negated_f(x):
        return -f(x)

    if x0 is None:
        x0 = np.mean(x, axis=1)  # Use midpoint of intervals as initial guess
    x0 = np.atleast_2d(x0)  # Ensure x0 is 2D
    if x0.shape[0] == 1:  # If only one initial guess is provided, use it for both min and max
        x0 = np.tile(x0, (2, 1))

    # Handle tol_loc
    if tol_loc is None:
        tol_min = None  # Use default tolerance
        tol_max = None  # Use default tolerance
    elif np.isscalar(tol_loc):  # If tol_loc is a scalar
        tol_min = tol_loc
        tol_max = tol_loc
    else:  # If tol_loc is an array
        tol_min = tol_loc[0]
        tol_max = tol_loc[1]

    #  Handle options_loc
    if options_loc is None:
        options_min = None  # Use default options
        options_max = None  # Use default options
    elif isinstance(options_loc, dict):  # If options_loc is a single dictionary
        options_min = options_loc
        options_max = options_loc
    else:  # If options_loc is a list of two dictionaries
        options_min = options_loc[0]
        options_max = options_loc[1]

    # Perform minimization and maximization
    min_y = minimize(f, x0=x0[0, :], method=method_loc,
                     bounds=bounds, tol=tol_min, options=options_min)
    max_y = minimize(negated_f, x0=x0[1, :], method=method_loc,
                     bounds=bounds, tol=tol_max, options=options_max)
    max_y.fun = -max_y.fun  # Correct the sign of the maximum value

    # Create an instance of the result class
    if results is None:
        results = Propagation_results()  # Create an instance of Propagation_results

    # Store ALL results in the results object with descriptions
    results.raw_data['min'] = np.append(results.raw_data['min'], {'x': min_y.x,
                                                                  'f': min_y.fun,
                                                                  'message': min_y.message,
                                                                  'niterations': min_y.nit,
                                                                  'nfevaluations': min_y.nfev,
                                                                  'final_simplex': min_y.final_simplex
                                                                  })

    results.raw_data['max'] = np.append(results.raw_data['max'], {
        'x': max_y.x,
        'f': max_y.fun,
        'message': max_y.message,
        'niterations': max_y.nit,
        'nfevaluations': max_y.nfev,
        'final_simplex': max_y.final_simplex
    })

    results.raw_data['bounds'] = np.array([min_y.fun, max_y.fun])

    return results

# # Example function
# f = lambda x: x[0] + x[1] + x[2]  # Example function
# x_bounds = np.array([[1, 2], [3, 4], [5, 6]])

# # Initial guess (same for min and max)
# x0 = np.array([1.5, 3.5, 5.5])

# # Different tolerances for min and max
# tol_loc = np.array([1e-4, 1e-6])

# # Different options for min and max
# options_loc = [
#     {'maxiter': 100},  # Options for minimization
#     {'maxiter': 1000}  # Options for maximization
#      ]

# # Perform optimization
# y = local_optimisation_method(x_bounds, f, x0=x0, tol_loc=tol_loc,
#                                 options_loc=options_loc)

# # Print the results
# y.print()



from __future__ import annotations
import json
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from .intervals.intervalOperators import wc_scalar_interval, make_vec_interval
from dataclasses import dataclass
from .intervals.number import Interval
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import matplotlib.path as mpath
from matplotlib.legend_handler import HandlerBase


def extend_ecdf(cdf):
    """add zero and one to the ecdf

    args:
        CDF_bundle
    """
    if cdf.probabilities[0] != 0:
        cdf.probabilities = np.insert(cdf.probabilities, 0, 0)
        cdf.quantiles = np.insert(cdf.quantiles, 0, cdf.quantiles[0])
    if cdf.probabilities[-1] != 1:
        cdf.probabilities = np.append(cdf.probabilities, 1)
        cdf.quantiles = np.append(cdf.quantiles, cdf.quantiles[-1])
    return cdf


def sorting(list1, list2):
    list1, list2 = (list(t) for t in zip(*sorted(zip(list1, list2))))
    return list1, list2


def reweighting(*masses):
    """reweight the masses to sum to 1"""
    masses = np.ravel(masses)
    return masses / masses.sum()


def uniform_reparameterisation(a, b):
    """reparameterise the uniform distribution to a, b"""
    #! incorrect in the case of Interval args
    a, b = wc_scalar_interval(a), wc_scalar_interval(b)
    return a, b - a


# TODO to test this high-performance version below
def find_nearest(array, value):
    """Find index/indices of nearest value(s) in `array` to each `value`.

    Efficient for both scalar and array inputs.
    """
    array = np.asarray(array)
    value = np.atleast_1d(value)

    # Compute distances using broadcasting
    diff = np.abs(array[None, :] - value[:, None])

    # Find index of minimum difference along axis 1
    indices = np.argmin(diff, axis=1)

    # Return scalar if input was scalar
    return indices[0] if np.isscalar(value) else indices


@mpl.rc_context({"text.usetex": True})
def plot_intervals(vec_interval: list[Interval], ax=None, **kwargs):
    """plot the intervals in a vectorised form
    args:
        vec_interval: vectorised interval objects
    """
    vec_interval = make_vec_interval(vec_interval)
    fig, ax = plt.subplots() if ax is None else (ax.figure, ax)
    for i, intl in enumerate(vec_interval):  # horizontally plot the interval
        ax.plot([intl.lo, intl.hi], [i, i], **kwargs)
    ax.margins(x=0.1, y=0.1)
    ax.set_yticks([])
    return ax


def read_json(file_name):
    f = open(file_name)
    data = json.load(f)
    return data


def is_increasing(arr):
    """check if 'arr' is increasing"""
    return np.all(np.diff(arr) >= 0)


class NotIncreasingError(Exception):
    pass


def condensation(bound, number):
    """a joint implementation for condensation"""
    if isinstance(bound, list | tuple):
        return condensation_bounds(bound, number)
    else:
        return condensation_bound(bound, number)


def condensation_bounds(bounds, number):
    """condense the bounds of number pbox

    args:
        number (int) : the number to be reduced
        bounds (list or tuple): the left and right bound to be reduced
    """
    b = bounds[0]

    if number > len(b):
        raise ValueError("Cannot sample more elements than exist in the list.")
    if len(bounds[0]) != len(bounds[1]):
        raise Exception("steps of two bounds are different")

    indices = np.linspace(0, len(b) - 1, number, dtype=int)

    l = np.array([bounds[0][i] for i in indices])
    r = np.array([bounds[1][i] for i in indices])
    return l, r


def condensation_bound(bound, number):
    """condense the bounds of number pbox

    args:
        number (int) : the number to be reduced
        bound (array-like): either the left or right bound to be reduced
    """

    if number > len(bound):
        raise ValueError("Cannot sample more elements than exist in the list.")

    indices = np.linspace(0, len(bound) - 1, number, dtype=int)

    new_bound = np.array([bound[i] for i in indices])
    return new_bound


def smooth_condensation(bounds, number=200):

    def smooth_ecdf(V, steps):

        m = len(V) - 1

        if m == 0:
            return np.repeat(V, steps)
        if steps == 1:
            return np.array([min(V), max(V)])

        d = 1 / m
        n = round(d * steps * 200)

        if n == 0:
            c = V
        else:
            c = []
            for i in range(m):
                v = V[i]
                w = V[i + 1]
                c.extend(np.linspace(start=v, stop=w, num=n))

        u = [c[round((len(c) - 1) * (k + 0) / (steps - 1))] for k in range(steps)]

        return np.array(u)

    l_smooth = smooth_ecdf(bounds[0], number)
    r_smooth = smooth_ecdf(bounds[1], number)
    return l_smooth, r_smooth


def equi_selection(arr, n):
    """draw n equidistant points from the array"""
    indices = np.linspace(0, len(arr) - 1, n, dtype=int)
    selected = arr[indices]
    return selected


# --- Reuse pbox rectangle key function ---
def create_colored_edge_box(x0, y0, width, height, linewidth=1):
    verts_top = [(x0, y0 + height), (x0 + width, y0 + height)]
    verts_left = [(x0, y0), (x0, y0 + height)]
    verts_bottom = [(x0, y0), (x0 + width, y0)]
    verts_right = [(x0 + width, y0), (x0 + width, y0 + height)]

    def make_patch(verts, color):
        path = mpath.Path(verts)
        return mpatches.PathPatch(
            path, edgecolor=color, facecolor="none", linewidth=linewidth
        )

    return [
        make_patch(verts_top, "green"),
        make_patch(verts_left, "green"),
        make_patch(verts_bottom, "blue"),
        make_patch(verts_right, "blue"),
    ]


# --- Custom pbox legend handler ---
class CustomEdgeRectHandler(HandlerBase):
    def create_artists(
        self, legend, orig_handle, xdescent, ydescent, width, height, fontsize, trans
    ):
        rect_patches = create_colored_edge_box(
            xdescent, ydescent, width, height, linewidth=1
        )
        for patch in rect_patches:
            patch.set_transform(trans)
        return rect_patches


import sys


def expose_functions_as_public(mapping, wrapper):
    """expose private functions as public APIs

    args:
        mapping (dict): a dictionary containing private function names mapped to public APIs
        wrapper (callable): a function that wraps the original functions (e.g., the decorator UNtoUN)

    note:
        the decorator which wraps the original function returning Pbox into returning UN, hence making the public UN API
    """
    # Get the module that called this function
    caller_globals = sys._getframe(1).f_globals
    for name, fn in mapping.items():
        caller_globals[name] = wrapper(fn)
